"""
From https://github.com/deepseek-ai/DeepSeek-VL2
"""

import math
from dataclasses import dataclass
from typing import Dict, List, Literal, Optional, Tuple

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from PIL import Image, ImageOps
from transformers import LlamaTokenizerFast
from transformers.processing_utils import ProcessorMixin

from .conversation import get_conv_template


def find_closest_aspect_ratio(aspect_ratio, target_ratios, width, height, image_size):
    best_ratio_diff = float("inf")
    best_ratio = (1, 1)
    area = width * height
    for ratio in target_ratios:
        target_aspect_ratio = ratio[0] / ratio[1]
        ratio_diff = abs(aspect_ratio - target_aspect_ratio)
        if ratio_diff < best_ratio_diff:
            best_ratio_diff = ratio_diff
            best_ratio = ratio
        elif ratio_diff == best_ratio_diff:
            if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                best_ratio = ratio
    # print(f'width: {width}, height: {height}, best_ratio: {best_ratio}')
    return best_ratio


def dynamic_preprocess(
    image, min_num=2, max_num=9, image_size=640, use_thumbnail=False
):
    orig_width, orig_height = image.size
    aspect_ratio = orig_width / orig_height

    # calculate the existing image aspect ratio
    target_ratios = set(
        (i, j)
        for n in range(min_num, max_num + 1)
        for i in range(1, n + 1)
        for j in range(1, n + 1)
        if i * j <= max_num and i * j >= min_num
    )
    # print(target_ratios)
    target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

    # find the closest aspect ratio to the target
    target_aspect_ratio = find_closest_aspect_ratio(
        aspect_ratio, target_ratios, orig_width, orig_height, image_size
    )

    # print(target_aspect_ratio)
    # calculate the target width and height
    target_width = image_size * target_aspect_ratio[0]
    target_height = image_size * target_aspect_ratio[1]
    blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

    # resize the image
    resized_img = image.resize((target_width, target_height))
    processed_images = []
    for i in range(blocks):
        box = (
            (i % (target_width // image_size)) * image_size,
            (i // (target_width // image_size)) * image_size,
            ((i % (target_width // image_size)) + 1) * image_size,
            ((i // (target_width // image_size)) + 1) * image_size,
        )
        # split the image
        split_img = resized_img.crop(box)
        processed_images.append(split_img)
    assert len(processed_images) == blocks
    if use_thumbnail and len(processed_images) != 1:
        thumbnail_img = image.resize((image_size, image_size))
        processed_images.append(thumbnail_img)
    return processed_images, target_aspect_ratio


class DictOutput(object):
    def keys(self):
        return self.__dict__.keys()

    def __getitem__(self, item):
        if isinstance(item, int):
            return list(self.__dict__.values())[item]
        if item not in self.__dict__:
            raise KeyError(item)
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value


@dataclass
class VLChatProcessorOutput(DictOutput):
    sft_format: str
    input_ids: mx.array
    target_ids: mx.array
    images: mx.array
    images_seq_mask: mx.array
    images_spatial_crop: mx.array
    num_image_tokens: List[int]

    def __len__(self):
        return len(self.input_ids)


@dataclass
class BatchCollateOutput(DictOutput):
    sft_format: List[str]
    input_ids: mx.array
    labels: mx.array
    images: mx.array
    attention_mask: mx.array
    images_seq_mask: mx.array
    images_spatial_crop: mx.array
    seq_lens: List[int]


class ImageTransform:
    def __init__(
        self,
        mean: Optional[Tuple[float, float, float]] = (0.5, 0.5, 0.5),
        std: Optional[Tuple[float, float, float]] = (0.5, 0.5, 0.5),
        normalize: bool = True,
    ):
        self.mean = mean
        self.std = std
        self.normalize = normalize

    def __call__(self, pil_img: Image.Image):
        # Convert PIL image to numpy array and normalize

        img = mx.array(np.array(pil_img)) / 255.0

        # Transpose from HWC to CHW format
        img = mx.transpose(img, [2, 0, 1])

        if self.normalize:
            mean = mx.array(self.mean).reshape(-1, 1, 1)
            std = mx.array(self.std).reshape(-1, 1, 1)
            img = (img - mean) / std

        return img


class DeepseekVLV2Processor(ProcessorMixin):
    tokenizer_class = ("LlamaTokenizer", "LlamaTokenizerFast")
    attributes = ["tokenizer"]

    def __init__(
        self,
        tokenizer: LlamaTokenizerFast,
        candidate_resolutions: Tuple[Tuple[int, int]],
        patch_size: int,
        downsample_ratio: int,
        image_mean: Tuple[float, float, float] = (0.5, 0.5, 0.5),
        image_std: Tuple[float, float, float] = (0.5, 0.5, 0.5),
        normalize: bool = True,
        image_token: str = "<image>",
        pad_token: str = "<｜▁pad▁｜>",
        add_special_token: bool = False,
        sft_format: str = "deepseek",
        mask_prompt: bool = True,
        ignore_id: int = -100,
        **kwargs,
    ):
        self.candidate_resolutions = candidate_resolutions
        self.image_size = candidate_resolutions[0][0]
        self.patch_size = patch_size
        self.image_mean = image_mean
        self.image_std = image_std
        self.normalize = normalize
        self.downsample_ratio = downsample_ratio

        self.image_transform = ImageTransform(
            mean=image_mean, std=image_std, normalize=normalize
        )
        self.tokenizer = tokenizer
        self.tokenizer.padding_side = "left"

        # Add special tokens
        if tokenizer.pad_token is None:
            self.tokenizer.add_special_tokens({"pad_token": pad_token})
        print(
            f"Add pad token = ['{pad_token}'] to the tokenizer\n"
            f"{pad_token}:{tokenizer.encode(pad_token, add_special_tokens=False)[0]}"
        )

        image_token_id = self.tokenizer.vocab.get(image_token)
        if image_token_id is None:
            special_tokens = [image_token]
            special_tokens_dict = {"additional_special_tokens": special_tokens}
            self.tokenizer.add_special_tokens(special_tokens_dict)
        self.image_token_id = self.tokenizer.vocab.get(image_token)
        print(
            f"Add image token = ['{image_token}'] to the tokenizer\n"
            f"{image_token}:{tokenizer.encode(image_token, add_special_tokens=False)[0]}"
        )

        # Add grounding-related tokens
        special_tokens = ["<|ref|>", "<|/ref|>", "<|det|>", "<|/det|>", "<|grounding|>"]
        special_tokens_dict = {"additional_special_tokens": special_tokens}
        self.tokenizer.add_special_tokens(special_tokens_dict)
        print("Added grounding-related tokens")

        # Add chat tokens
        special_tokens = ["<|User|>", "<|Assistant|>"]
        special_tokens_dict = {"additional_special_tokens": special_tokens}
        self.tokenizer.add_special_tokens(special_tokens_dict)
        print("Added chat tokens")

        self.image_token = image_token
        self.pad_token = pad_token
        self.add_special_token = add_special_token
        self.sft_format = sft_format
        self.mask_prompt = mask_prompt
        self.ignore_id = ignore_id

        super().__init__(tokenizer, **kwargs)

        # Add chat template
        self.chat_template = kwargs.pop("chat_template", self.default_chat_template)

    @property
    def default_chat_template(self):
        return (
            "{% for message in messages %}"
            "{% if message['role'] == 'user' %}"
            "{% elif message['role'] == 'assistant' %}{% endif %}"
            "{{message['content']}} "
            "{% endfor %}"
            "{% if add_generation_prompt %}{% endif %}"
        )

    @property
    def bos_id(self):
        return self.tokenizer.bos_token_id

    @property
    def eos_id(self):
        return self.tokenizer.eos_token_id

    @property
    def pad_id(self):
        return self.tokenizer.pad_token_id

    def encode(self, text: str, bos: bool = True, eos: bool = False):
        t = self.tokenizer.encode(text, add_special_tokens=False)

        if bos:
            t = [self.bos_id] + t
        if eos:
            t = t + [self.eos_id]

        return t

    def decode(self, t: List[int], **kwargs) -> str:
        return self.tokenizer.decode(t, **kwargs)

    def process_one(
        self,
        prompt: str = None,
        conversations: List[Dict[str, str]] = None,
        images: List[Image.Image] = None,
        apply_sft_format: bool = False,
        inference_mode: bool = True,
        system_prompt: str = "",
        **kwargs,
    ):
        assert (
            prompt is None or conversations is None
        ), "prompt and conversations cannot be used at the same time."

        if apply_sft_format:
            sft_format = self.format_prompts(
                prompts=prompt, sft_format=self.sft_format, system_prompt=system_prompt
            )
        else:
            sft_format = prompt
        (
            tokenized_str,
            images_list,
            images_seq_mask,
            images_spatial_crop,
            num_image_tokens,
        ) = self.tokenize_with_images(sft_format, images, cropping=len(images) <= 2)
        masked_tokenized_str = []
        for token_index in tokenized_str:
            if token_index != self.image_token_id:
                masked_tokenized_str.append(token_index)
            else:
                masked_tokenized_str.append(self.ignore_id)

        input_ids = mx.array(tokenized_str)
        target_ids = mx.array(masked_tokenized_str)
        images_seq_mask = mx.array(images_seq_mask)

        # Set ignored indices
        target_ids = mx.where(
            (input_ids < 0) | (input_ids == self.image_token_id),
            self.ignore_id,
            target_ids,
        )
        input_ids = mx.where(input_ids < 0, self.pad_id, input_ids)

        if inference_mode:
            input_ids = input_ids[:-1]
            target_ids = target_ids[:-1]
            images_seq_mask = images_seq_mask[:-1]

        return {
            "input_ids": input_ids[None, :],
            "attention_mask": input_ids != self.pad_id,
            "labels": target_ids,
            "images": images_list,
            "images_seq_mask": images_seq_mask[None, ...],
            "images_spatial_crop": images_spatial_crop,
            "num_image_tokens": num_image_tokens,
        }

    def pad_sequence(self, sequences, padding_value):
        # Get max length of sequences
        max_len = max(len(seq) for seq in sequences)

        # Pad each sequence to max length
        padded_seqs = []
        for seq in sequences:
            pad_length = max_len - len(seq)
            if pad_length > 0:
                padding = mx.full((pad_length,), padding_value)
                padded_seq = mx.concatenate([seq, padding])
            else:
                padded_seq = seq
            padded_seqs.append(padded_seq)

        return mx.stack(padded_seqs)

    def tokenize_with_images(
        self,
        conversation: str,
        images: List[Image.Image],
        base_size: int = 1024,
        image_size: int = 640,
        cropping: bool = True,
    ):

        patch_size = 16
        downsample_ratio = 4
        valid_img_tokens = 0
        ratio = 1

        image_draw = images[0].copy()

        w, h = image_draw.size

        ratio = 1 - ((max(w, h) - min(w, h)) / (max(w, h)))

        """Tokenize text with <image> tags."""
        assert conversation.count(self.image_token) == len(images)
        text_splits = conversation.split(self.image_token)

        images_list, images_crop_list, images_seq_mask = [], [], []
        tokenized_str = []
        images_spatial_crop = []
        for text_sep, image in zip(text_splits, images):

            tokenized_sep = self.encode(text_sep, bos=False, eos=False)
            tokenized_str += tokenized_sep
            images_seq_mask += [False] * len(tokenized_sep)

            if cropping:

                if image.size[0] <= 640 and image.size[1] <= 640:
                    crop_ratio = [1, 1]

                else:
                    if cropping:
                        # best_width, best_height = select_best_resolution(image.size, self.candidate_resolutions)
                        images_crop_raw, crop_ratio = dynamic_preprocess(image)

                    else:
                        # best_width, best_height = self.image_size, self.image_size
                        crop_ratio = [1, 1]

                """process the global view"""
                # image = image.resize((base_size, base_size))
                global_view = ImageOps.pad(
                    image,
                    (base_size, base_size),
                    color=tuple(int(x * 255) for x in self.image_transform.mean),
                )

                if base_size == 1024:
                    valid_img_tokens += int(256 * ratio)
                elif base_size == 1280:
                    valid_img_tokens += int(400 * ratio)
                # elif base_size == 640:
                #     valid_img_tokens += int(100 * ratio)

                images_list.append(
                    self.image_transform(global_view).astype(mx.bfloat16)
                )

                # global_view_tensor = image_transform(global_view).to(torch.bfloat16)

                width_crop_num, height_crop_num = crop_ratio

                images_spatial_crop.append([width_crop_num, height_crop_num])

                if width_crop_num > 1 or height_crop_num > 1:
                    """process the local views"""

                    for i in range(len(images_crop_raw)):
                        images_crop_list.append(
                            self.image_transform(images_crop_raw[i]).astype(mx.bfloat16)
                        )

                if image_size == 640:
                    valid_img_tokens += len(images_crop_list) * 100

                num_queries = math.ceil((image_size // patch_size) / downsample_ratio)
                num_queries_base = math.ceil(
                    (base_size // patch_size) / downsample_ratio
                )

                """add image tokens"""

                tokenized_image = (
                    [self.image_token_id] * num_queries_base + [self.image_token_id]
                ) * num_queries_base
                tokenized_image += [self.image_token_id]
                if width_crop_num > 1 or height_crop_num > 1:
                    tokenized_image += (
                        [self.image_token_id] * (num_queries * width_crop_num)
                        + [self.image_token_id]
                    ) * (num_queries * height_crop_num)
                tokenized_str += tokenized_image
                images_seq_mask += [True] * len(tokenized_image)
                # num_image_tokens.append(len(tokenized_image))

            else:
                # best_width, best_height = self.image_size, self.image_size
                # print(image.size, (best_width, best_height)) # check the select_best_resolutions func

                """process the global view"""
                if image_size <= 640:
                    print("directly resize")
                    image = image.resize((image_size, image_size))
                # else:
                global_view = ImageOps.pad(
                    image,
                    (image_size, image_size),
                    color=tuple(int(x * 255) for x in self.image_transform.mean),
                )
                images_list.append(
                    self.image_transform(global_view).astype(mx.bfloat16)
                )

                if base_size == 1024:
                    valid_img_tokens += int(256 * ratio)
                elif base_size == 1280:
                    valid_img_tokens += int(400 * ratio)
                elif base_size == 640:
                    valid_img_tokens += int(100 * 1)
                elif base_size == 512:
                    valid_img_tokens += int(64 * 1)

                width_crop_num, height_crop_num = 1, 1

                images_spatial_crop.append([width_crop_num, height_crop_num])

                """add image tokens"""
                num_queries = math.ceil((image_size // patch_size) / downsample_ratio)

                tokenized_image = (
                    [self.image_token_id] * num_queries + [self.image_token_id]
                ) * num_queries
                tokenized_image += [self.image_token_id]
                # tokenized_image += ([self.image_token_id] * (num_queries * width_crop_num) + [self.image_token_id]) * (
                #             num_queries * height_crop_num)
                tokenized_str += tokenized_image
                images_seq_mask += [True] * len(tokenized_image)

        tokenized_sep = self.encode(text_splits[-1], bos=False, eos=False)
        tokenized_str += tokenized_sep
        images_seq_mask += [False] * len(tokenized_sep)

        """add the bos tokens"""
        bos_id = 0
        tokenized_str = [bos_id] + tokenized_str
        images_seq_mask = [False] + images_seq_mask

        images_seq_mask = mx.array(images_seq_mask)

        if len(images_list) == 0:
            images_ori = mx.zeros((1, 3, image_size, image_size))
            images_spatial_crop = mx.zeros((1, 2))
            images_crop = mx.zeros((1, 3, base_size, base_size))

        else:
            images_ori = mx.stack(images_list, axis=0)
            images_spatial_crop = mx.array(images_spatial_crop)
            if images_crop_list:
                images_crop = mx.stack(images_crop_list, axis=0)
            else:
                images_crop = mx.zeros((1, 3, base_size, base_size))

        assert len(tokenized_str) == len(
            images_seq_mask
        ), f"tokenize_with_images func: tokenized_str's length {len(tokenized_str)} is not equal to imags_seq_mask's length {len(images_seq_mask)}"

        return (
            tokenized_str,
            [images_crop, images_ori],
            images_seq_mask,
            images_spatial_crop,
            valid_img_tokens,
        )

    def __call__(
        self,
        *,
        text: str = None,
        images: List[Image.Image] = None,
        apply_sft_format: bool = False,
        inference_mode: bool = True,
        system_prompt: str = "",
        **kwargs,
    ):
        """

        Args:
            prompt (str): the formatted prompt;
            conversations (List[Dict]): conversations with a list of messages;
            images (List[ImageType]): the list of images;
            apply_sft_format (bool): if prompt is not None, then apply the SFT format to prompt;
                if conversations is not None, then it will always apply the SFT format to conversations;
            force_batchify (bool): force batchify the inputs;
            inference_mode (bool): if True, then remove the last eos token;
            system_prompt (str): the system prompt;
            **kwargs:

        Returns:
            outputs (BaseProcessorOutput): the output of the processor,
                - input_ids (torch.LongTensor): [N + image tokens]
                - images (torch.FloatTensor): [n_images, 3, H, W]
                - image_id (int): the id of the image token
                - num_image_tokens (List[int]): the number of image tokens
        """

        prepare = self.process_one(
            prompt=text,
            images=images,
            apply_sft_format=apply_sft_format,
            inference_mode=inference_mode,
            system_prompt=system_prompt,
        )

        return prepare



import mlx.core as mx
import numpy as np


def gaussian_blur_axis(image, sigma, axis):
    """
    Applies a 1D Gaussian blur along the given axis.
    This version works for arrays with any number of dimensions.
    """
    radius = int(3 * sigma)
    if radius < 1:
        return image
    x = mx.arange(-radius, radius + 1)
    kernel = mx.exp(-(x**2) / (2 * sigma**2))
    kernel = kernel / mx.sum(kernel)

    # MLX doesn't have a direct apply_along_axis equivalent,
    # so we'll implement the convolution differently based on the axis

    # Helper function to apply 1D convolution along specific axis
    def conv_1d(array, kernel, axis):
        # Reshape kernel to broadcast along the right dimensions
        kernel_shape = [1] * image.ndim
        kernel_shape[axis] = len(kernel)
        kernel_reshaped = kernel.reshape(kernel_shape)

        # Pad the array
        pad_width = [(0, 0)] * image.ndim
        pad_width[axis] = (radius, radius)
        padded = mx.pad(array, pad_width, mode="edge")

        # Perform convolution via sliding window sum
        result = mx.zeros_like(array)
        slices = [slice(None)] * padded.ndim

        for i in range(2 * radius + 1):
            slices[axis] = slice(i, i + array.shape[axis])
            result = result + padded[tuple(slices)] * kernel_reshaped

        return result

    return conv_1d(image, kernel, axis)


def bilinear_interpolate(image, new_height, new_width, align_corners=False):
    """
    Performs bilinear interpolation on an array whose spatial dimensions are the first two.
    It supports extra dimensions (e.g. channels or batch dimensions that have been moved to the trailing axes).
    """
    # image is assumed to have shape (H, W, ...) where H and W are spatial dimensions.
    H_in, W_in = image.shape[0], image.shape[1]

    # Compute sampling positions in the input image.
    if new_height == 1:
        row_positions = mx.array([0.0])
    else:
        if align_corners:
            row_positions = mx.linspace(0, H_in - 1, new_height)
        else:
            row_positions = (mx.arange(new_height) + 0.5) * H_in / new_height - 0.5

    if new_width == 1:
        col_positions = mx.array([0.0])
    else:
        if align_corners:
            col_positions = mx.linspace(0, W_in - 1, new_width)
        else:
            col_positions = (mx.arange(new_width) + 0.5) * W_in / new_width - 0.5

    # Compute floor and ceil indices.
    row_floor = mx.floor(row_positions).astype(mx.int32)
    col_floor = mx.floor(col_positions).astype(mx.int32)
    row_ceil = row_floor + 1
    col_ceil = col_floor + 1

    row_floor = mx.clip(row_floor, 0, H_in - 1)
    row_ceil = mx.clip(row_ceil, 0, H_in - 1)
    col_floor = mx.clip(col_floor, 0, W_in - 1)
    col_ceil = mx.clip(col_ceil, 0, W_in - 1)

    row_weight = row_positions - row_floor  # shape (new_height,)
    col_weight = col_positions - col_floor  # shape (new_width,)

    # Use advanced indexing for gather operations
    # Create meshgrid for coordinates
    row_floor_grid, col_floor_grid = mx.meshgrid(row_floor, col_floor, indexing="ij")
    row_ceil_grid, col_floor_grid = mx.meshgrid(row_ceil, col_floor, indexing="ij")
    row_floor_grid, col_ceil_grid = mx.meshgrid(row_floor, col_ceil, indexing="ij")
    row_ceil_grid, col_ceil_grid = mx.meshgrid(row_ceil, col_ceil, indexing="ij")

    # Gather the four surrounding pixels using take_along_axis
    # For higher dimensional arrays, we'll need to reshape and broadcast
    extra_dims = image.ndim - 2

    def gather_pixels(row_indices, col_indices):
        # Flatten the spatial dimensions for gathering
        flat_indices = row_indices * W_in + col_indices
        flat_image = mx.reshape(image, (-1,) + image.shape[2:])
        # Gather and reshape back
        gathered = mx.take(flat_image, flat_indices.reshape(-1), axis=0)
        return mx.reshape(gathered, (new_height, new_width) + image.shape[2:])

    top_left = gather_pixels(row_floor_grid, col_floor_grid)
    top_right = gather_pixels(row_floor_grid, col_ceil_grid)
    bottom_left = gather_pixels(row_ceil_grid, col_floor_grid)
    bottom_right = gather_pixels(row_ceil_grid, col_ceil_grid)

    # Expand the weights to have shape (new_height, new_width, *[1]*extra_dims)
    r_weight = row_weight.reshape(new_height, 1, *([1] * extra_dims))
    c_weight = col_weight.reshape(1, new_width, *([1] * extra_dims))

    # Perform bilinear interpolation.
    result = (
        (1 - r_weight) * (1 - c_weight) * top_left
        + (1 - r_weight) * c_weight * top_right
        + r_weight * (1 - c_weight) * bottom_left
        + r_weight * c_weight * bottom_right
    )
    return result


def resize_bilinear(image, new_size, align_corners=False, antialias=True):
    """
    Resizes an image (or embedding tensor) to new_size=(new_height, new_width)
    using bilinear interpolation with MLX.

    Supports:
      - 2D: (H, W)
      - 3D: (H, W, C)
      - 4D: (B, C, H, W)  (assumed for typical image batches)
    """
    new_height, new_width = new_size

    # Convert numpy arrays to MLX arrays if needed
    if isinstance(image, np.ndarray):
        image = mx.array(image)

    if image.ndim == 2 or image.ndim == 3:
        # Assume spatial dims are the first two.
        resized = image
        H_in, W_in = image.shape[:2]
        if antialias:
            if new_height < H_in:
                scale_y = new_height / H_in
                sigma_y = (1 / scale_y - 1) / 2.0  # heuristic
                if sigma_y > 0:
                    resized = gaussian_blur_axis(resized, sigma_y, axis=0)
            if new_width < W_in:
                scale_x = new_width / W_in
                sigma_x = (1 / scale_x - 1) / 2.0
                if sigma_x > 0:
                    resized = gaussian_blur_axis(resized, sigma_x, axis=1)
        resized = bilinear_interpolate(
            resized, new_height, new_width, align_corners=align_corners
        )
        return resized

    elif image.ndim == 4:
        # Assume shape is (B, C, H, W) (typical PyTorch/MLX format).
        B, C, H_in, W_in = image.shape
        # Permute to bring spatial dims to the front: (H, W, B, C)
        image_perm = mx.transpose(image, (2, 3, 0, 1))
        resized = image_perm
        if antialias:
            if new_height < H_in:
                scale_y = new_height / H_in
                sigma_y = (1 / scale_y - 1) / 2.0
                if sigma_y > 0:
                    resized = gaussian_blur_axis(resized, sigma_y, axis=0)
            if new_width < W_in:
                scale_x = new_width / W_in
                sigma_x = (1 / scale_x - 1) / 2.0
                if sigma_x > 0:
                    resized = gaussian_blur_axis(resized, sigma_x, axis=1)
        resized = bilinear_interpolate(
            resized, new_height, new_width, align_corners=align_corners
        )
        # Permute back to (B, C, new_height, new_width)
        resized = mx.transpose(resized, (2, 3, 0, 1))
        return resized

    else:
        raise ValueError("Unsupported image dimensions.")


#



import glob
import inspect
import json
import re
from pathlib import Path
from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from huggingface_hub import snapshot_download
from transformers import AutoConfig

from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class MLP(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        input_size = config.vision_config.hidden_size * (config.scale_factor**2)
        output_size = config.text_config.hidden_size
        self.proj = nn.Linear(input_size, output_size, bias=False)

    def __call__(self, x):
        return self.proj(x)


class Idefics3Connector(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.scale_factor = config.scale_factor
        self.modality_projection = MLP(config)

    def pixel_shuffle(self, x, scale_factor=2):
        bsz, seq, embed_dim = x.shape
        height = width = int(seq**0.5)
        x = x.reshape(bsz, height, width, embed_dim)
        x = x.reshape(bsz, height, int(width / scale_factor), embed_dim * scale_factor)
        x = x.transpose(0, 2, 1, 3)
        x = x.reshape(
            bsz,
            int(width / scale_factor),
            int(height / scale_factor),
            embed_dim * (scale_factor**2),
        )
        x = x.transpose(0, 2, 1, 3)
        x = x.reshape(bsz, int(seq / (scale_factor**2)), embed_dim * (scale_factor**2))
        return x

    def __call__(self, image_hidden_states):
        image_hidden_states = self.pixel_shuffle(image_hidden_states, self.scale_factor)
        image_hidden_states = self.modality_projection(image_hidden_states)
        return image_hidden_states


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.model_type = config.model_type
        self.config = config

        self.vision_model = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.connector = Idefics3Connector(config)

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        pixel_attention_mask: Optional[mx.array] = None,
    ):
        if pixel_values is None:
            return self.language_model.embed_tokens(input_ids)

        inputs_embeds = self.language_model.embed_tokens(input_ids)

        pooler_output, embeddings, hidden_state = self.vision_model(
            pixel_values[0].transpose(0, 2, 3, 1), output_hidden_states=True
        )

        image_features = pooler_output.astype(pixel_values.dtype)
        image_features = self.connector(image_features)

        final_inputs_embeds = self._prepare_inputs_for_multimodal(
            image_features, inputs_embeds, input_ids
        )
        return final_inputs_embeds

    def _prepare_inputs_for_multimodal(self, image_features, inputs_embeds, input_ids):
        image_token_index = self.config.image_token_index

        # Positions of <image> tokens in input_ids, assuming batch size is 1
        image_positions = np.where(input_ids == image_token_index)[1].tolist()

        num_images, _, vision_hidden_size = image_features.shape

        reshaped_image_hidden_states = image_features.reshape(-1, vision_hidden_size)

        # cast to the dtype of the input_embeds to support quantized models
        reshaped_image_hidden_states = reshaped_image_hidden_states.astype(
            inputs_embeds.dtype
        )
        inputs_embeds[:, image_positions, :] = reshaped_image_hidden_states

        return inputs_embeds

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        cache=None,
        **kwargs,
    ):
        input_embeddings = self.get_input_embeddings(input_ids, pixel_values)
        logits = self.language_model(
            inputs=input_ids, cache=cache, inputs_embeds=input_embeddings
        )
        return logits

    def sanitize(self, weights):
        weights = {
            (
                f"{k.split('.', 1)[1]}"
                if re.match(r"^model\.", k)
                else (f"language_model.{k}" if re.match(r"^lm_head\.", k) else k)
            ): v
            for k, v in weights.items()
        }

        weights = {
            (
                f"language_model.{k.split('.', 1)[1]}"
                if re.match(
                    r"^text_model\.",
                    k,
                )
                else k
            ): v
            for k, v in weights.items()
        }

        return weights



import glob
import inspect
import json
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from huggingface_hub import snapshot_download

from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.language_model = LanguageModel(config.text_config)
        self.vision_tower = VisionModel(config.vision_config)

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        mask: mx.array,
        cache=None,
        **kwargs,
    ) -> Dict[str, Union[mx.array, List[Tuple[mx.array, mx.array]]]]:
        if input_ids.ndim == 1:
            input_ids = input_ids[None, :]

        batch_size, seq_len = input_ids.shape

        image_input_idx = kwargs.get("image_input_idx", None)
        image_masks = kwargs.get("image_masks", None)

        if pixel_values is not None:
            assert (
                image_masks is not None and image_input_idx is not None
            ), "image_masks and image_input_idx must be provided when images are given"

            dtype = self.vision_tower.image_vit.patch_embedding.weight.dtype
            pixel_values = pixel_values.astype(dtype)

            # Process images
            if pixel_values.ndim == 3:
                pixel_values = mx.expand_dims(pixel_values, 0)
                image_masks = (
                    mx.expand_dims(image_masks, 0) if image_masks is not None else None
                )
                image_input_idx = (
                    mx.expand_dims(image_input_idx, 0)
                    if image_input_idx is not None
                    else None
                )

            image_features, cls_embed = self.vision_tower(pixel_values, image_masks)

            # Insert image features into the input embeddings
            num_image, num_patch = image_features.shape[1:3]

            assert image_input_idx.shape == (
                batch_size,
                num_image,
                num_patch,
            ), f"image_input_idx.shape: {image_input_idx.shape}, expected: {(batch_size, num_image, num_patch)}"

            # Insert image features into the input embeddings
            image_features = image_features.reshape(
                batch_size, num_image * num_patch, -1
            )
            image_input_idx = image_input_idx.reshape(batch_size, num_image * num_patch)

            valid = np.where(image_input_idx >= 0)[0].tolist()
            batch_idx = mx.arange(batch_size)
            batch_idx = mx.tile(batch_idx[:, None], [1, image_features.shape[1]])

            input_embeddings = self.language_model.model.wte(input_ids)
            input_embeddings[
                batch_idx[valid], image_input_idx[valid]
            ] += image_features[valid]
        else:
            input_embeddings = None

        # Forward pass through the language model
        logits = self.language_model(
            input_ids,
            inputs_embeds=input_embeddings,
            mask=mask,
            cache=cache,
        )

        return logits

    def sanitize(self, weights):
        def transform_key(key):
            if "model.transformer" in key:
                key = key.replace("model.transformer", "language_model.model")
            if "model.vision_backbone" in key:
                key = key.replace("model.vision_backbone", "vision_tower")
            return key

        return {transform_key(k): v for k, v in weights.items()}



from functools import partial
from typing import Optional, Tuple, Type

import mlx.core as mx
import mlx.nn as nn


def get_abs_pos_sam(abs_pos, tgt_size):
    """Interpolate absolute positional embeddings to target size."""
    dtype = abs_pos.dtype
    src_size = abs_pos.shape[1]

    if src_size != tgt_size:
        # Transpose to (B, C, H, W) for interpolation
        old_pos_embed = abs_pos.transpose(0, 3, 1, 2)
        old_pos_embed = old_pos_embed.astype(mx.float32)

        # Bicubic interpolation
        from ..kernels import bicubic_interpolate

        new_pos_embed = bicubic_interpolate(
            old_pos_embed, size=(tgt_size, tgt_size), antialias=True
        ).astype(dtype)

        # Transpose back to (B, H, W, C)
        new_pos_embed = new_pos_embed.transpose(0, 2, 3, 1)
        return new_pos_embed
    else:
        return abs_pos


class MLPBlock(nn.Module):
    """MLP block with GELU activation."""

    def __init__(
        self,
        embedding_dim: int,
        mlp_dim: int,
        act: Type[nn.Module] = nn.GELU,
    ) -> None:
        super().__init__()
        self.lin1 = nn.Linear(embedding_dim, mlp_dim)
        self.lin2 = nn.Linear(mlp_dim, embedding_dim)
        self.act = act()

    def __call__(self, x: mx.array) -> mx.array:
        return self.lin2(self.act(self.lin1(x)))


class Attention(nn.Module):
    """Multi-head Attention block with relative position embeddings."""

    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = True,
        use_rel_pos: bool = False,
        input_size: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Args:
            dim (int): Number of input channels.
            num_heads (int): Number of attention heads.
            qkv_bias (bool): If True, add a learnable bias to query, key, value.
            use_rel_pos (bool): If True, add relative positional embeddings to the attention map.
            input_size (tuple(int, int) or None): Input resolution for calculating the relative
                positional parameter size.
        """
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)

        self.use_rel_pos = use_rel_pos
        if self.use_rel_pos:
            assert (
                input_size is not None
            ), "Input size must be provided if using relative positional encoding."
            # Initialize relative positional embeddings
            self.rel_pos_h = mx.zeros((2 * input_size[0] - 1, head_dim))
            self.rel_pos_w = mx.zeros((2 * input_size[1] - 1, head_dim))

    def __call__(self, x: mx.array) -> mx.array:
        B, H, W, _ = x.shape

        # QKV projection and reshape
        qkv = (
            self.qkv(x)
            .reshape(B, H * W, 3, self.num_heads, -1)
            .transpose(2, 0, 3, 1, 4)
        )

        # Separate q, k, v
        qkv_reshaped = qkv.reshape(3, B * self.num_heads, H * W, -1)
        q, k, v = qkv_reshaped[0], qkv_reshaped[1], qkv_reshaped[2]

        # Compute relative positional embeddings if needed
        rel_h, rel_w = None, None
        if self.use_rel_pos:
            rel_h, rel_w = add_decomposed_rel_pos(
                q, self.rel_pos_h, self.rel_pos_w, (H, W), (H, W)
            )

        # Reshape for attention
        q = q.reshape(B, self.num_heads, H * W, -1)
        k = k.reshape(B, self.num_heads, H * W, -1)
        v = v.reshape(B, self.num_heads, H * W, -1)

        # Apply scaled dot product attention
        if self.use_rel_pos:
            rel_h = rel_h.reshape(
                B, self.num_heads, rel_h.shape[1], rel_h.shape[2], rel_h.shape[3]
            )
            rel_w = rel_w.reshape(
                B, self.num_heads, rel_w.shape[1], rel_w.shape[2], rel_w.shape[3]
            )
            attn_bias = (rel_h + rel_w).reshape(
                B, self.num_heads, rel_h.shape[2], rel_h.shape[3] * rel_w.shape[4]
            )
            x = mx.fast.scaled_dot_product_attention(
                q, k, v, scale=self.scale, mask=attn_bias
            )
        else:
            x = mx.fast.scaled_dot_product_attention(q, k, v, scale=self.scale)

        # Reshape output
        x = (
            x.reshape(B, self.num_heads, H, W, -1)
            .transpose(0, 2, 3, 1, 4)
            .reshape(B, H, W, -1)
        )

        x = self.proj(x)
        return x


class Block(nn.Module):
    """Transformer blocks with support of window attention and residual propagation."""

    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        act_layer: Type[nn.Module] = nn.GELU,
        use_rel_pos: bool = False,
        window_size: int = 0,
        input_size: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Args:
            dim (int): Number of input channels.
            num_heads (int): Number of attention heads in each ViT block.
            mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            qkv_bias (bool): If True, add a learnable bias to query, key, value.
            norm_layer (nn.Module): Normalization layer.
            act_layer (nn.Module): Activation layer.
            use_rel_pos (bool): If True, add relative positional embeddings to the attention map.
            window_size (int): Window size for window attention blocks. If it equals 0, then
                use global attention.
            input_size (tuple(int, int) or None): Input resolution for calculating the relative
                positional parameter size.
        """
        super().__init__()
        self.norm1 = norm_layer(dim, eps=1e-6)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            use_rel_pos=use_rel_pos,
            input_size=input_size if window_size == 0 else (window_size, window_size),
        )

        self.norm2 = norm_layer(dim, eps=1e-6)
        self.mlp = MLPBlock(
            embedding_dim=dim, mlp_dim=int(dim * mlp_ratio), act=act_layer
        )

        self.window_size = window_size

    def __call__(self, x: mx.array) -> mx.array:
        shortcut = x
        x = self.norm1(x)

        # Window partition
        if self.window_size > 0:
            H, W = x.shape[1], x.shape[2]
            x, pad_hw = window_partition(x, self.window_size)

        x = self.attn(x)

        # Reverse window partition
        if self.window_size > 0:
            x = window_unpartition(x, self.window_size, pad_hw, (H, W))

        x = shortcut + x
        x = x + self.mlp(self.norm2(x))

        return x


class PatchEmbed(nn.Module):
    """Image to Patch Embedding."""

    def __init__(
        self,
        kernel_size: Tuple[int, int] = (16, 16),
        stride: Tuple[int, int] = (16, 16),
        in_chans: int = 3,
        embed_dim: int = 768,
    ) -> None:
        """
        Args:
            kernel_size (Tuple): kernel size of the projection layer.
            stride (Tuple): stride of the projection layer.
            in_chans (int): Number of input image channels.
            embed_dim (int): Patch embedding dimension.
        """
        super().__init__()
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=kernel_size, stride=stride
        )

    def __call__(self, x: mx.array) -> mx.array:
        x = self.proj(x)
        return x


class SAMEncoder(nn.Module):
    """Vision Transformer encoder based on SAM architecture."""

    def __init__(
        self,
        img_size: int = 1024,
        patch_size: int = 16,
        in_chans: int = 3,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        out_chans: int = 256,
        qkv_bias: bool = True,
        norm_layer: Type[nn.Module] = nn.LayerNorm,
        act_layer: Type[nn.Module] = nn.GELU,
        use_abs_pos: bool = True,
        use_rel_pos: bool = True,
        window_size: int = 14,
        global_attn_indexes: Tuple[int, ...] = (2, 5, 8, 11),
    ) -> None:
        """
        Args:
            img_size (int): Input image size.
            patch_size (int): Patch size.
            in_chans (int): Number of input image channels.
            embed_dim (int): Patch embedding dimension.
            depth (int): Depth of ViT.
            num_heads (int): Number of attention heads in each ViT block.
            mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            out_chans (int): Output channels for neck.
            qkv_bias (bool): If True, add a learnable bias to query, key, value.
            norm_layer (nn.Module): Normalization layer.
            act_layer (nn.Module): Activation layer.
            use_abs_pos (bool): If True, use absolute positional embeddings.
            use_rel_pos (bool): If True, add relative positional embeddings to the attention map.
            window_size (int): Window size for window attention blocks.
            global_attn_indexes (tuple): Indexes for blocks using global attention.
        """
        super().__init__()
        self.img_size = img_size

        self.patch_embed = PatchEmbed(
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            in_chans=in_chans,
            embed_dim=embed_dim,
        )

        self.use_abs_pos = use_abs_pos
        if use_abs_pos:
            # Initialize absolute positional embedding with pretrain image size
            self.pos_embed = mx.zeros(
                (1, img_size // patch_size, img_size // patch_size, embed_dim)
            )

        self.blocks = []
        for i in range(depth):
            block = Block(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                norm_layer=norm_layer,
                act_layer=act_layer,
                use_rel_pos=use_rel_pos,
                window_size=window_size if i not in global_attn_indexes else 0,
                input_size=(img_size // patch_size, img_size // patch_size),
            )
            self.blocks.append(block)

        # Neck layers for output processing
        self.neck = [
            nn.Conv2d(embed_dim, out_chans, kernel_size=1, bias=False),
            nn.LayerNorm(out_chans, eps=1e-6),
            nn.Conv2d(out_chans, out_chans, kernel_size=3, padding=1, bias=False),
            nn.LayerNorm(out_chans, eps=1e-6),
        ]

        # Additional downsampling layers
        self.net_2 = nn.Conv2d(256, 512, kernel_size=3, stride=2, padding=1, bias=False)
        self.net_3 = nn.Conv2d(
            512, 1024, kernel_size=3, stride=2, padding=1, bias=False
        )

    def __call__(self, x: mx.array) -> mx.array:
        # Patch embedding
        x = self.patch_embed(x)

        # Add positional embeddings
        if self.use_abs_pos:
            x = x + get_abs_pos_sam(self.pos_embed, x.shape[1])

        # Apply transformer blocks
        for blk in self.blocks:
            x = blk(x)

        # Apply neck layers
        for n in self.neck:
            x = n(x)

        # Additional downsampling
        x = self.net_2(x)
        x = self.net_3(x)

        return x


# Utility functions


def window_partition(x: mx.array, window_size: int) -> Tuple[mx.array, Tuple[int, int]]:
    """
    Partition into non-overlapping windows with padding if needed.

    Args:
        x (mx.array): input tokens with [B, H, W, C].
        window_size (int): window size.

    Returns:
        windows: windows after partition with [B * num_windows, window_size, window_size, C].
        (Hp, Wp): padded height and width before partition
    """
    B, H, W, C = x.shape

    pad_h = (window_size - H % window_size) % window_size
    pad_w = (window_size - W % window_size) % window_size

    if pad_h > 0 or pad_w > 0:
        x = mx.pad(x, [(0, 0), (0, pad_h), (0, pad_w), (0, 0)])

    Hp, Wp = H + pad_h, W + pad_w

    x = x.reshape(B, Hp // window_size, window_size, Wp // window_size, window_size, C)
    windows = x.transpose(0, 1, 3, 2, 4, 5).reshape(-1, window_size, window_size, C)

    return windows, (Hp, Wp)


def window_unpartition(
    windows: mx.array,  # FIXED: Changed from np.ndarray to mx.array
    window_size: int,
    pad_hw: Tuple[int, int],
    hw: Tuple[int, int],
) -> mx.array:  # FIXED: Changed return type from implicit to mx.array
    """
    Window unpartition into original sequences and removing padding.

    Args:
        windows (mx.array): input tokens with [B * num_windows, window_size, window_size, C].
        window_size (int): window size.
        pad_hw (Tuple): padded height and width (Hp, Wp).
        hw (Tuple): original height and width (H, W) before padding.

    Returns:
        x: unpartitioned sequences with [B, H, W, C].
    """
    Hp, Wp = pad_hw
    H, W = hw
    B = windows.shape[0] // (Hp * Wp // window_size // window_size)

    x = windows.reshape(
        B, Hp // window_size, Wp // window_size, window_size, window_size, -1
    )
    x = x.transpose(0, 1, 3, 2, 4, 5).reshape(B, Hp, Wp, -1)

    if Hp > H or Wp > W:
        x = x[:, :H, :W, :]

    return x


def get_rel_pos(q_size: int, k_size: int, rel_pos: mx.array) -> mx.array:
    """
    Get relative positional embeddings according to the relative positions of
    query and key sizes.

    Args:
        q_size (int): size of query q.
        k_size (int): size of key k.
        rel_pos (mx.array): relative position embeddings (L, C).

    Returns:
        Extracted positional embeddings according to relative positions.
    """
    max_rel_dist = int(2 * max(q_size, k_size) - 1)

    # Interpolate rel pos if needed
    if rel_pos.shape[0] != max_rel_dist:
        dtype = rel_pos.dtype
        rel_pos = rel_pos.astype(mx.float32)
        rel_pos_resized = rel_pos.reshape(1, rel_pos.shape[0], -1).transpose(0, 2, 1)

        # Linear interpolation
        scale = rel_pos_resized.shape[2] / max_rel_dist
        indices = mx.arange(max_rel_dist, dtype=mx.float32) * scale
        idx_floor = mx.floor(indices).astype(mx.int32)
        idx_ceil = mx.minimum(idx_floor + 1, rel_pos_resized.shape[2] - 1)
        weight = indices - idx_floor.astype(mx.float32)

        rel_pos_resized = (
            mx.take(rel_pos_resized, idx_floor, axis=2) * (1 - weight)
            + mx.take(rel_pos_resized, idx_ceil, axis=2) * weight
        ).astype(dtype)

        rel_pos_resized = rel_pos_resized.reshape(-1, max_rel_dist).transpose(1, 0)
    else:
        rel_pos_resized = rel_pos

    # Scale the coords with short length if shapes for q and k are different
    q_coords = mx.arange(q_size, dtype=mx.float32)[:, None] * max(k_size / q_size, 1.0)
    k_coords = mx.arange(k_size, dtype=mx.float32)[None, :] * max(q_size / k_size, 1.0)
    relative_coords = (q_coords - k_coords) + (k_size - 1) * max(q_size / k_size, 1.0)

    return rel_pos_resized[relative_coords.astype(mx.int32)]


def add_decomposed_rel_pos(
    q: mx.array,  # FIXED: Changed from np.ndarray to mx.array
    rel_pos_h: mx.array,  # FIXED: Changed from np.ndarray to mx.array
    rel_pos_w: mx.array,  # FIXED: Changed from np.ndarray to mx.array
    q_size: Tuple[int, int],
    k_size: Tuple[int, int],
) -> Tuple[mx.array, mx.array]:  # FIXED: Added explicit return type
    """
    Calculate decomposed Relative Positional Embeddings.

    Args:
        q (mx.array): query q in the attention layer with shape (B, q_h * q_w, C).
        rel_pos_h (mx.array): relative position embeddings (Lh, C) for height axis.
        rel_pos_w (mx.array): relative position embeddings (Lw, C) for width axis.
        q_size (Tuple): spatial sequence size of query q with (q_h, q_w).
        k_size (Tuple): spatial sequence size of key k with (k_h, k_w).

    Returns:
        Tuple of (rel_h, rel_w): relative position biases for height and width.
    """
    q_h, q_w = q_size
    k_h, k_w = k_size

    Rh = get_rel_pos(q_h, k_h, rel_pos_h)
    Rw = get_rel_pos(q_w, k_w, rel_pos_w)

    B, _, dim = q.shape
    r_q = q.reshape(B, q_h, q_w, dim)

    rel_h = mx.einsum("bhwc,hkc->bhwk", r_q, Rh)
    rel_w = mx.einsum("bhwc,wkc->bhwk", r_q, Rw)
    rel_h = rel_h[..., None]
    rel_w = rel_w[..., None, :]
    rel_h = rel_h.reshape(B, q_h * q_w, k_h, 1)
    rel_w = rel_w.reshape(B, q_h * q_w, 1, k_w)

    return rel_h, rel_w



import inspect
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, List, Optional, Union

import mlx.core as mx
import mlx.nn as nn

from ..base import (
    LanguageModelOutput,
    create_attention_mask,
    scaled_dot_product_attention,
)
from ..cache import KVCache, RotatingKVCache
from .config import TextConfig


class RMSNorm(nn.Module):
    def __init__(self, dims: int, eps: float = 1e-5):
        super().__init__()
        self.weight = mx.ones((dims,))
        self.eps = eps

    def __call__(self, x):
        return mx.fast.rms_norm(x, 1.0 + self.weight, self.eps)


class Attention(nn.Module):
    def __init__(self, config: TextConfig, layer_idx: int):
        super().__init__()

        dim = config.hidden_size
        self.n_heads = n_heads = config.num_attention_heads
        self.n_kv_heads = n_kv_heads = config.num_key_value_heads
        self.repeats = n_heads // n_kv_heads
        self.head_dim = head_dim = config.head_dim
        self.layer_idx = layer_idx

        self.scale = config.query_pre_attn_scalar**-0.5

        self.q_proj = nn.Linear(dim, n_heads * head_dim, bias=False)
        self.k_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=False)
        self.v_proj = nn.Linear(dim, n_kv_heads * head_dim, bias=False)
        self.o_proj = nn.Linear(n_heads * head_dim, dim, bias=False)

        self.q_norm = RMSNorm(dims=head_dim, eps=config.rms_norm_eps)
        self.k_norm = RMSNorm(dims=head_dim, eps=config.rms_norm_eps)
        self.is_sliding = (layer_idx + 1) % config.sliding_window_pattern != 0

        self.rope = nn.RoPE(
            head_dim,
            traditional=config.rope_traditional,
            base=(
                config.rope_local_base_freq
                if self.is_sliding
                else config.rope_global_base_freq
            ),
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Any] = None,
    ) -> mx.array:
        B, L, _ = x.shape
        queries, keys, values = self.q_proj(x), self.k_proj(x), self.v_proj(x)
        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)

        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        queries = self.q_norm(queries)
        keys = self.k_norm(keys)

        if cache is not None:
            queries = self.rope(queries, offset=cache.offset)
            keys = self.rope(keys, offset=cache.offset)
            keys, values = cache.update_and_fetch(keys, values)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        # Sliding window
        if mask is not None and isinstance(mask, mx.array):
            if mask.shape[-1] != keys.shape[-2]:
                mask = mask[..., -keys.shape[-2] :]

        output = scaled_dot_product_attention(
            queries, keys, values, cache, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)
        return self.o_proj(output)


class MLP(nn.Module):
    def __init__(self, dim, hidden_dim):
        super().__init__()
        self.gate_proj = nn.Linear(dim, hidden_dim, bias=False)
        self.down_proj = nn.Linear(hidden_dim, dim, bias=False)
        self.up_proj = nn.Linear(dim, hidden_dim, bias=False)

    def __call__(self, x) -> mx.array:
        # This should not be GELU approx, jax.nn.gelu
        return self.down_proj(nn.gelu_approx(self.gate_proj(x)) * self.up_proj(x))


@partial(mx.compile, shapeless=True)
def clip_residual(x, y=None):
    bound = mx.finfo(mx.float16).max
    if y is None:
        if x.dtype == mx.float16:
            return mx.clip(x.astype(mx.float32), -bound, bound).astype(mx.float16)
        else:
            return x

    if x.dtype != mx.float16:
        return x + y

    return mx.clip(x.astype(mx.float32) + y.astype(mx.float32), -bound, bound).astype(
        mx.float16
    )


class TransformerBlock(nn.Module):
    def __init__(self, config: TextConfig, layer_idx: int):
        super().__init__()
        self.num_attention_heads = config.num_attention_heads
        self.hidden_size = config.hidden_size
        self.self_attn = Attention(config, layer_idx)
        self.mlp = MLP(config.hidden_size, config.intermediate_size)
        self.input_layernorm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)
        self.post_attention_layernorm = RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.pre_feedforward_layernorm = RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )
        self.post_feedforward_layernorm = RMSNorm(
            config.hidden_size, eps=config.rms_norm_eps
        )

    def __call__(
        self,
        x: mx.array,
        mask: Optional[mx.array] = None,
        cache: Optional[Any] = None,
    ) -> mx.array:

        # Clip the input to avoid overflow in float16
        # Float16 has a max value of 65504. When values exceed this limit, they become inf.
        # Example: If x contains 70000.0 in float16, it becomes inf, causing gradient issues.
        # We upcast to float32 for operations that might exceed the limit, then clip and
        # convert back to float16 to maintain numerical stability.

        # Clip input to avoid overflow in float16
        x = clip_residual(x)

        # Self-attention block
        r = self.self_attn(self.input_layernorm(x), mask, cache)
        h = self.post_attention_layernorm(r)

        # Add residual connection with overflow protection for float16
        h = clip_residual(x + h)

        # MLP block
        r = self.mlp(self.pre_feedforward_layernorm(h))
        out = self.post_feedforward_layernorm(r)

        # Add residual connection with overflow protection for float16
        out = clip_residual(h + out)

        return out


class Gemma3Model(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.vocab_size = config.vocab_size
        self.num_hidden_layers = config.num_hidden_layers
        assert self.vocab_size > 0
        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size)
        self.layers = [
            TransformerBlock(config=config, layer_idx=layer_idx)
            for layer_idx in range(config.num_hidden_layers)
        ]
        self.norm = RMSNorm(config.hidden_size, eps=config.rms_norm_eps)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: mx.array = None,
        mask: mx.array = None,
        cache=None,
    ):
        if inputs_embeds is None:
            h = self.embed_tokens(inputs)
        else:
            h = inputs_embeds

        h *= mx.array(self.config.hidden_size**0.5, mx.bfloat16).astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        if mask is None:
            j = self.config.sliding_window_pattern
            full_mask = create_attention_mask(h, cache[j - 1 : j])
            sliding_window_mask = create_attention_mask(h, cache)

        for i, (layer, c) in enumerate(zip(self.layers, cache)):
            is_global = (
                i % self.config.sliding_window_pattern
                == self.config.sliding_window_pattern - 1
            )

            local_mask = mask
            if mask is None and is_global:
                local_mask = full_mask
            elif mask is None:
                local_mask = sliding_window_mask

            h = layer(h, local_mask, c)

        return self.norm(h)


class LanguageModel(nn.Module):
    def __init__(self, config: TextConfig):
        super().__init__()
        self.config = config
        self.model_type = config.model_type
        self.model = Gemma3Model(config)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

    def __call__(
        self,
        inputs: mx.array,
        inputs_embeds: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
    ):
        out = self.model(inputs, inputs_embeds=inputs_embeds, mask=mask, cache=cache)
        out = self.lm_head(out)
        return LanguageModelOutput(logits=out)

    def sanitize(self, weights):
        if "lm_head.weight" not in weights:
            weights["language_model.lm_head.weight"] = weights[
                "language_model.model.embed_tokens.weight"
            ]
        return {
            k: v for k, v in weights.items() if "self_attn.rotary_emb.inv_freq" not in k
        }

    @property
    def layers(self):
        return self.model.layers

    @property
    def head_dim(self):
        return self.config.head_dim

    @property
    def n_kv_heads(self):
        return self.config.num_key_value_heads

    def make_cache(self):
        caches = []
        for i in range(self.config.num_hidden_layers):
            if (
                i % self.config.sliding_window_pattern
                == self.config.sliding_window_pattern - 1
            ):
                caches.append(KVCache())
            else:
                caches.append(
                    RotatingKVCache(
                        max_size=self.config.sliding_window,
                        keep=0,
                    )
                )
        return caches



import inspect
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

from ..base import BaseModelConfig


@dataclass
class TextConfig(BaseModelConfig):
    model_type: str = "deepseek_v2"
    vocab_size: int = 102400
    hidden_size: int = 1280
    intermediate_size: int = 6848
    moe_intermediate_size: int = 896
    num_hidden_layers: int = 30
    num_attention_heads: int = 32
    num_key_value_heads: int = 32
    n_shared_experts: Optional[int] = 2
    n_routed_experts: Optional[int] = 64
    routed_scaling_factor: float = 1.0
    kv_lora_rank: int = 512
    q_lora_rank: int = 1536
    qk_rope_head_dim: int = 0
    v_head_dim: int = 128
    qk_nope_head_dim: int = 0
    topk_method: str = "greedy"
    n_group: Optional[int] = 1
    topk_group: Optional[int] = 1
    num_experts_per_tok: Optional[int] = 6
    moe_layer_freq: int = 1
    first_k_dense_replace: int = 0
    max_position_embeddings: int = 2048
    rms_norm_eps: float = 1e-6
    rope_theta: float = 10000.0
    rope_traditional: bool = False
    rope_scaling: Dict = None
    attention_bias: bool = False
    scoring_func: str = "softmax"
    attn_type: str = "DeepseekV2Attention"

    def __post_init__(self):
        if self.qk_nope_head_dim == 0:
            self.attn_type = "LlamaAttention"

        if self.num_key_value_heads is None:
            self.num_key_value_heads = self.num_attention_heads


@dataclass
class VisionConfig(BaseModelConfig):
    model_type: str
    layers: int = 24
    width: int = 1152
    hidden_size: int = 1024
    intermediate_size: int = 4096
    num_attention_heads: int = 16
    image_size: int = 224
    patch_size: int = 14
    num_channels: int = 3
    layer_norm_eps: float = 1e-6
    mlp_ratio: float = 3.7362
    cls: str = None
    params: dict = None


@dataclass
class MLPConfig(BaseModelConfig):
    hidden_size: int
    intermediate_size: int
    hidden_act: str = "gelu"


@dataclass
class ProjectorConfig(BaseModelConfig):
    projector_type: str = "linear"
    input_dim: int = 2048
    n_embed: int = 1280
    depth: int = 2
    mlp_ratio: int = 1
    downsample_ratio: int = 2
    token_pooling: bool = False


@dataclass
class SAMViTConfig(BaseModelConfig):
    image_size: Union[Tuple[int, int], int] = 1024
    width: int = 768
    layers: int = 12
    heads: int = 12
    patch_size: int = 16
    window_size: int = 14
    prompt_embed_dim: int = 256
    global_attn_indexes: Union[List[int], Tuple[int]] = (2, 5, 8, 11)
    downsample_channels: Union[List[int], Tuple[int]] = (512, 1024)


@dataclass
class ModelConfig(BaseModelConfig):
    text_config: TextConfig
    vision_config: VisionConfig
    projector_config: ProjectorConfig
    model_type: str
    ignore_index: int = -100
    image_token_index: int = 128815
    vision_feature_select_strategy: str = "default"
    select_layer: int = -1
    pad_id: int = 100001
    num_image_tokens: int = 576
    vocab_size: int = 32000
    tile_tag: str = "2D"
    global_view_pos: str = "head"
    eos_token_id: Optional[List[int]] = None
    quantization: Optional[Dict] = None

    @classmethod
    def from_dict(cls, params):
        if "language_config" in params:
            params["text_config"] = params["language_config"]
            del params["language_config"]

        return cls(
            text_config=TextConfig.from_dict(params["text_config"]),
            vision_config=VisionConfig.from_dict(params["vision_config"]),
            projector_config=ProjectorConfig.from_dict(params["projector_config"]),
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
                and k not in ["text_config", "vision_config", "projector_config"]
            },
        )


@dataclass
class Conversation:
    """A class that represents a conversation."""

    system: str
    roles: List[str]
    messages: List[List[str]]
    offset: int
    sep_style: int
    sep: str
    sep2: str
    version: str = "Unknown"


@dataclass
class VLChatProcessorOutput:
    """
    Output of the VL chat processor.
    """

    sft_format: str
    input_ids: List[int]
    pixel_values: List
    num_image_tokens: List[int]
    image_grid_thw: List[List[int]]
    image_sizes: Optional[List[List[int]]] = None
    videos: Optional[List] = None
    aspect_ratio_ids: Optional[List[int]] = None
    aspect_ratio_mask: Optional[List[List[int]]] = None
    cross_attention_mask: Optional[List[List[List[int]]]] = None
    attention_mask: Optional[List[int]] = None
    labels: Optional[List[int]] = None


@dataclass
class BatchCollateOutput:
    input_ids: List
    labels: List
    attention_mask: List
    pixel_values: List



import math
from typing import Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from transformers import AutoProcessor

from .config import ModelConfig, ProjectorConfig, SAMViTConfig
from .language import LanguageModel
from .processing_deepseekocr import DeepseekVLV2Processor
from .sam import SAMEncoder
from .vision import VisionModel

AutoProcessor.register("deepseekocr", DeepseekVLV2Processor)


class MlpProjector(nn.Module):
    def __init__(self, config: ProjectorConfig):
        super().__init__()
        self.config = config

        if config.projector_config.projector_type == "linear":
            modules = nn.Linear(
                config.projector_config.input_dim, config.projector_config.n_embed
            )

        elif config.projector_config.projector_type == "downsample_mlp_gelu":
            mlp_depth = config.projector_config.depth
            mlp_ratio = config.projector_config.mlp_ratio
            modules = [
                nn.Linear(
                    config.projector_config.input_dim
                    * config.projector_config.downsample_ratio
                    * config.projector_config.downsample_ratio,
                    config.projector_config.n_embed * mlp_ratio,
                )
            ]
            for _ in range(1, mlp_depth - 1):
                modules.append(nn.GELU())
                modules.append(
                    nn.Linear(
                        config.projector_config.n_embed * mlp_ratio,
                        config.projector_config.n_embed * mlp_ratio,
                    )
                )
            modules.append(nn.GELU())
            modules.append(
                nn.Linear(
                    config.projector_config.n_embed * mlp_ratio,
                    config.projector_config.n_embed,
                )
            )
        else:
            raise ValueError(
                f"Unknown projector type: {config.projector_config.projector_type}"
            )

        self.layers = modules

    def __call__(self, x):
        if self.config.projector_config.projector_type == "downsample_mlp_gelu":
            bs, hw, input_dim = x.shape
            h = w = int(math.sqrt(hw))

            # Compute padding
            pad = (
                0
                if h % self.config.projector_config.downsample_ratio == 0
                else self.config.projector_config.downsample_ratio
                - h % self.config.projector_config.downsample_ratio
            )

            x = mx.reshape(x, (bs, h, w, input_dim))
            if pad > 0:
                x = mx.pad(x, [(0, 0), (0, pad), (0, pad), (0, 0)], constant_values=0)

            x = mx.transpose(x, (0, 3, 1, 2))  # B, C, H, W

            # Manual implementation of unfold for downsampling
            h_pad, w_pad = x.shape[2], x.shape[3]
            ds = self.config.projector_config.downsample_ratio
            patches = []

            for i in range(0, h_pad - ds + 1, ds):
                for j in range(0, w_pad - ds + 1, ds):
                    patch = x[:, :, i : i + ds, j : j + ds]
                    patches.append(mx.reshape(patch, (bs, -1)))

            x = mx.stack(patches, axis=1)  # B, N_patches, C*ds*ds

        if self.config.projector_config.projector_type == "linear":
            x = self.layers(x)
        else:
            for layer in self.layers:
                x = layer(x)
        return x


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_model = VisionModel(config.vision_config)
        sam_config = SAMViTConfig()
        self.sam_model = SAMEncoder(
            img_size=sam_config.image_size,
            patch_size=sam_config.patch_size,
            embed_dim=sam_config.width,
            depth=sam_config.layers,
            num_heads=sam_config.heads,
            window_size=sam_config.window_size,
            global_attn_indexes=sam_config.global_attn_indexes,
        )
        self.language_model = LanguageModel(config.text_config)
        self.projector = MlpProjector(config)
        self.vision_feature_layer = config.select_layer
        self.vision_feature_select_strategy = config.vision_feature_select_strategy

        self.tile_tag = config.tile_tag
        self.global_view_pos = config.global_view_pos
        # 用于format image token sequence的特殊token
        embed_std = 1 / mx.sqrt(
            mx.array(config.projector_config.n_embed, dtype=mx.float32)
        )

        if self.tile_tag == "2D":

            # <|view_separator|>, <|\n|>
            self.image_newline = mx.array(
                mx.random.normal((config.projector_config.n_embed,)) * embed_std
            )
            # fix the typo: view_seperater
            self.view_separator = mx.array(
                mx.random.normal((config.projector_config.n_embed,)) * embed_std
            )
        else:
            raise ValueError(
                f"Only 2D tile_tag is supported currently, got: {self.tile_tag}"
            )

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        images_spatial_crop: Optional[mx.array] = None,
        images_seq_mask: Optional[mx.array] = None,
    ):
        input_embeds = self.language_model.model.embed_tokens(input_ids)

        if pixel_values is None:
            return input_embeds

        if (
            self.sam_model is not None
            and (input_ids.shape[1] != 1 or self.training)
            and mx.sum(pixel_values[0][1]).item() != 0
        ):

            idx = 0

            for crop_shape in images_spatial_crop.tolist():
                images_in_this_batch = []
                patches = pixel_values[0]
                image_ori = pixel_values[1]
                if mx.sum(patches).item() != 0:
                    crop_flag = 1
                    local_features_1 = self.sam_model(patches.transpose(0, 2, 3, 1))

                    local_features_2 = self.vision_model(
                        patches.transpose(0, 2, 3, 1), patch_embeds=local_features_1
                    )

                    local_features = mx.concatenate(
                        (
                            local_features_2[:, 1:],
                            local_features_1.flatten(start_axis=1, end_axis=2),
                        ),
                        axis=-1,
                    )

                    local_features = self.projector(local_features)

                    global_features_1 = self.sam_model(image_ori.transpose(0, 2, 3, 1))
                    global_features_2 = self.vision_model(
                        image_ori.transpose(0, 2, 3, 1), global_features_1
                    )

                    global_features = mx.concatenate(
                        (
                            global_features_2[:, 1:],
                            global_features_1.flatten(start_axis=1, end_axis=2),
                        ),
                        axis=-1,
                    )
                    global_features = self.projector(global_features)

                    print("=====================")
                    print("BASE: ", global_features.shape)
                    print("PATCHES: ", local_features.shape)
                    print("=====================")

                    _, hw, n_dim = global_features.shape
                    h = w = int(hw**0.5)

                    _2, hw2, n_dim2 = local_features.shape
                    h2 = w2 = int(hw2**0.5)

                    width_crop_num, height_crop_num = (
                        crop_shape[0],
                        crop_shape[1],
                    )

                    global_features = global_features.reshape(h, w, n_dim)

                    global_features = mx.concatenate(
                        [
                            global_features,
                            mx.broadcast_to(
                                self.image_newline[None, None, :], (h, 1, n_dim)
                            ),
                        ],
                        axis=1,
                    )

                    global_features = global_features.reshape(-1, n_dim)

                    local_features = (
                        local_features.reshape(
                            height_crop_num, width_crop_num, h2, w2, n_dim2
                        )
                        .transpose(0, 2, 1, 3, 4)
                        .reshape(height_crop_num * h2, width_crop_num * w2, n_dim2)
                    )
                    local_features = mx.concatenate(
                        [
                            local_features,
                            mx.broadcast_to(
                                self.image_newline[None, None, :],
                                (height_crop_num * h2, 1, n_dim2),
                            ),
                        ],
                        axis=1,
                    )
                    local_features = local_features.reshape(-1, n_dim2)

                    global_local_features = mx.concatenate(
                        [local_features, global_features, self.view_separator[None, :]],
                        axis=0,
                    )

                else:
                    global_features_1 = self.sam_model(image_ori.transpose(0, 2, 3, 1))
                    global_features_2 = self.vision_model(
                        image_ori.transpose(0, 2, 3, 1), global_features_1
                    )
                    global_features = mx.concatenate(
                        (
                            global_features_2[:, 1:],
                            global_features_1.flatten(start_axis=1, end_axis=2),
                        ),
                        axis=-1,
                    )
                    global_features = self.projector(global_features)
                    print("=====================")
                    print("BASE: ", global_features.shape)
                    print("NO PATCHES")
                    print("=====================")
                    _, hw, n_dim = global_features.shape
                    h = w = int(hw**0.5)

                    global_features = global_features.reshape(h, w, n_dim)

                    global_features = mx.concatenate(
                        [
                            global_features,
                            mx.broadcast_to(
                                self.image_newline[None, None, :], (h, 1, n_dim)
                            ),
                        ],
                        axis=1,
                    )

                    global_features = global_features.reshape(-1, n_dim)

                    global_local_features = mx.concatenate(
                        [global_features, self.view_separator[None, :]], axis=0
                    )

                images_in_this_batch.append(global_local_features)

                if images_in_this_batch:
                    images_in_this_batch = mx.concatenate(images_in_this_batch, axis=0)
                    # Find positions where images should be placed
                    image_indices = np.where(images_seq_mask[idx])[0].tolist()
                    # Directly assign the image features to those positions
                    input_embeds[idx, image_indices] = images_in_this_batch

                idx += 1

        return input_embeds

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: Optional[mx.array] = None,
        mask: Optional[mx.array] = None,
        cache=None,
        **kwargs,
    ):

        images_spatial_crop = kwargs.get("images_spatial_crop", None)
        images_seq_mask = kwargs.get("images_seq_mask", None)

        input_embeddings = self.get_input_embeddings(
            input_ids, pixel_values, images_spatial_crop, images_seq_mask
        )

        logits = self.language_model(
            input_ids, cache=cache, inputs_embeds=input_embeddings
        )
        return logits

    @staticmethod
    def sanitize(weights):
        def transform_key(key):
            if "model.layers" in key and "language_model" not in key:
                key = key.replace("model.layers", "language_model.model.layers")

            if "model.embed_tokens" in key and "language_model" not in key:
                key = key.replace(
                    "model.embed_tokens", "language_model.model.embed_tokens"
                )

            if "model.norm" in key and "language_model" not in key:
                key = key.replace("model.norm", "language_model.model.norm")

            if "model.vision_model" in key:
                key = key.replace("model.vision_model", "vision_model")

            if "model.sam_model" in key:
                key = key.replace("model.sam_model", "sam_model")

            if "model.projector" in key:
                key = key.replace("model.projector", "projector")

            if "model.view_seperator" in key:
                key = key.replace("model.view_seperator", "view_separator")

            if "model.image_newline" in key:
                key = key.replace("model.image_newline", "image_newline")

            if "lm_head.weight" in key and "language_model" not in key:
                key = key.replace("lm_head.weight", "language_model.lm_head.weight")

            return key

        return {transform_key(k): v for k, v in weights.items()}



import json
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from typing import Union

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from mlx.utils import tree_flatten, tree_map


def get_prompt(model_type, processor, conversation):
    if model_type == "paligemma":
        return conversation

    if "chat_template" in processor.__dict__.keys():
        prompt = processor.apply_chat_template(
            conversation,
            tokenize=False,
            add_generation_prompt=False,
        )
    elif "tokenizer" in processor.__dict__.keys():
        prompt = processor.tokenizer.apply_chat_template(
            conversation,
            tokenize=False,
            add_generation_prompt=False,
        )

    return prompt


class Dataset:
    def __init__(
        self,
        hf_dataset,
        config,
        processor,
        image_processor=None,
        take=None,
        split=None,
        image_resize_shape=None,
    ):
        if split is not None:
            self.dataset = hf_dataset[split]
        else:
            self.dataset = hf_dataset
        if take is not None:
            self.dataset = self.dataset.take(take)
        self.processor = processor
        self.config = config
        self.image_processor = image_processor
        self.image_resize_shape = image_resize_shape

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        from mlx_vlm.utils import prepare_inputs

        item = self.dataset[idx]

        images = item.get("images", item.get("image", None))
        conversations = item.get("messages", item.get("conversations"))
        if images in (None, "", []):
            images = []
        prompts = []

        if isinstance(conversations, list) and isinstance(conversations[0], list):
            for conversation in conversations:
                if self.config["model_type"] == "pixtral":
                    conversation = [json.loads(i) for i in conversation]
                    if len(conversations) > 1:
                        warnings.warn(
                            "Pixtral batch processing is not supported yet. Set batch size to 1."
                        )

                prompt = get_prompt(
                    self.config["model_type"], self.processor, conversation
                )
                prompts.append(prompt)

        else:
            if self.config["model_type"] == "pixtral":
                conversations = [json.loads(i) for i in conversations]
            prompt = get_prompt(
                self.config["model_type"], self.processor, conversations
            )
            prompts.append(prompt)

        image_token_index = self.config["image_token_index"]

        inputs = prepare_inputs(
            processor=self.processor,
            images=images,
            audio=None,
            prompts=prompts,
            image_token_index=image_token_index,
            resize_shape=self.image_resize_shape,
        )
        input_ids = inputs["input_ids"]
        pixel_values = inputs["pixel_values"]
        mask = inputs["attention_mask"]
        kwargs = {
            k: v
            for k, v in inputs.items()
            if k not in ["input_ids", "pixel_values", "attention_mask"]
        }

        if mask is None:
            mask = mx.ones_like(input_ids)

        return {
            "pixel_values": pixel_values,
            "input_ids": input_ids,
            "attention_mask": mask,
            **kwargs,
        }


def grad_checkpoint(layer):
    """
    Update all instances of type(layer) to use gradient checkpointing.
    """
    fn = type(layer).__call__

    def checkpointed_fn(model, *args, **kwargs):
        def inner_fn(params, *args, **kwargs):
            model.update(params)
            return fn(model, *args, **kwargs)

        return mx.checkpoint(inner_fn)(model.trainable_parameters(), *args, **kwargs)

    type(layer).__call__ = checkpointed_fn


@dataclass
class TrainingArgs:
    batch_size: int = field(default=4, metadata={"help": "Minibatch size."})
    iters: int = field(default=100, metadata={"help": "Iterations to train for."})
    val_batches: int = field(
        default=25,
        metadata={
            "help": "Number of validation batches, -1 uses the entire validation set."
        },
    )
    steps_per_report: int = field(
        default=10,
        metadata={"help": "Number of training steps between loss reporting."},
    )
    steps_per_eval: int = field(
        default=200, metadata={"help": "Number of training steps between validations."}
    )
    steps_per_save: int = field(
        default=100, metadata={"help": "Save the model every number steps"}
    )
    max_seq_length: int = field(
        default=2048, metadata={"help": "Maximum sequence length."}
    )
    adapter_file: str = field(
        default="adapters.safetensors",
        metadata={"help": "Save/load path for the trained adapter weights."},
    )
    grad_checkpoint: bool = field(
        default=False,
        metadata={"help": "Use gradient checkpointing to reduce memory use."},
    )


def default_loss(model, inputs, targets, lengths):
    logits = model(inputs)
    logits = logits.astype(mx.float32)

    length_mask = mx.arange(inputs.shape[1])[None, :] < lengths[:, None]

    ce = nn.losses.cross_entropy(logits, targets) * length_mask
    ntoks = length_mask.sum()
    ce = ce.sum() / ntoks

    return ce, ntoks


class Trainer:
    def __init__(
        self,
        model,
        optimizer,
        train_on_completions=False,
        assistant_id=77091,
        clip_gradients=None,
    ):
        self.model = model
        self.optimizer = optimizer
        self.train_on_completions = train_on_completions
        self.assistant_id = assistant_id
        self.clip_gradients = clip_gradients

    def loss_fn(self, model, batch):
        pixel_values = batch["pixel_values"]
        input_ids = batch["input_ids"]
        attention_mask = batch["attention_mask"]
        lengths = mx.sum(attention_mask, axis=1)
        labels = input_ids[:, 1:]

        batch_size, seq_length = input_ids.shape

        if self.train_on_completions:
            weight_mask = mx.ones_like(attention_mask)

            assistant_response_index = np.where(input_ids == self.assistant_id)[1]
            range_matrix = mx.repeat(
                mx.expand_dims(mx.arange(seq_length), 0), batch_size, axis=0
            )
            assistant_mask = range_matrix <= mx.array(assistant_response_index).reshape(
                -1, 1
            )
            # Apply the mask to weight_mask
            weight_mask = mx.where(
                assistant_mask, mx.zeros_like(weight_mask), weight_mask
            )[:, 1:]
        else:
            weight_mask = None

        input_ids = input_ids[:, :-1]

        kwargs = {
            k: v
            for k, v in batch.items()
            if k not in ["input_ids", "pixel_values", "attention_mask"]
        }

        # Forward pass
        outputs = model(input_ids, pixel_values, attention_mask, **kwargs)

        # Cast to float32
        logits = outputs.logits.astype(mx.float32)

        # Ensure logits and labels have the same sequence length
        def align_logits_with_labels(logits, labels):
            if logits.shape[1] < labels.shape[1]:
                pad_length = labels.shape[1] - logits.shape[1]
                pad_width = ((0, 0), (0, pad_length), (0, 0))
                return mx.pad(logits, pad_width, mode="constant", constant_values=-100)
            elif logits.shape[1] > labels.shape[1]:
                return logits[:, -labels.shape[1] :, :]
            return logits

        logits = align_logits_with_labels(logits, labels)

        length_mask = mx.arange(input_ids.shape[1])[None, :] < lengths[:, None]

        # Compute loss only on non-padded tokens
        ce = (
            nn.losses.cross_entropy(
                logits,
                labels,
                weights=weight_mask,
            )
            * length_mask
        )
        ntoks = length_mask.sum()
        ce = ce.sum() / ntoks

        return ce

    def train_step(self, batch):
        loss_and_grad_fn = nn.value_and_grad(self.model, self.loss_fn)
        loss, grads = loss_and_grad_fn(self.model, batch)

        # Add gradient clipping
        if self.clip_gradients is not None:
            grads = tree_map(
                lambda g: mx.clip(g, -self.clip_gradients, self.clip_gradients), grads
            )

        self.optimizer.update(self.model, grads)

        return loss

    @mx.compile
    def train_epoch(self, dataloader):
        total_loss = 0
        for batch in dataloader:
            loss = self.train_step(batch)
            mx.eval(self.model, self.optimizer.state)
            total_loss += loss
        return total_loss / len(dataloader)


def save_adapter(
    model: nn.Module,
    adapter_file: Union[str, Path],
):
    path = Path(adapter_file)
    if hasattr(model.config, "lora"):
        with open(path.parent / "adapter_config.json", "w") as f:
            json.dump(model.config.lora, f)
    flattened_tree = tree_flatten(model.trainable_parameters())
    mx.save_safetensors(str(adapter_file), dict(flattened_tree))



from collections.abc import Sequence
from math import sqrt
from typing import Dict, List, Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from mlx_vlm.models.gemma3n.config import (
    EdgeResidualConfig,
    MultiQueryAttentionBlockConfig,
    UniversalInvertedResidualConfig,
    VisionConfig,
)

from ..kernels import bicubic_interpolate, nearest_interpolate


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/models/mobilenetv5.py#L24
class MobileNetV5MultiScaleFusionAdapter(nn.Module):
    """Multi-layer fusion token adapter.
    Attributes:
      out_filters: The number of output filters.
      output_resolution: The output resolution.
      activation: The activation function.
      expansion_ratio: The expansion ratio.
      upsampling_interpolation: The upsampling interpolation.
      use_layer_scale: Whether to use layer scale.
      layer_scale_init_value: The initial value of the layer scale.
      skip_projection: Whether to skip the projection.
      name: The name of the module.
      upsize: The upsampling fn.
      downsize: The downsampling fn.
    """

    def __init__(
        self,
        in_chs: List[int],
        out_chs: int,
        output_resolution: int,
        expansion_ratio: float = 2.0,
        interpolation_mode: str = "nearest",
        use_layer_scale: bool = False,
        layer_scale_init_value: float = 1e-5,
        noskip: bool = True,
    ):
        super().__init__()
        self.in_channels = sum(in_chs) if isinstance(in_chs, Sequence) else in_chs
        self.out_channels = out_chs
        self.output_resolution = to_2tuple(output_resolution)
        self.expansion_ratio = expansion_ratio
        self.interpolation_mode = interpolation_mode
        self.use_layer_scale = use_layer_scale
        self.layer_scale_init_value = layer_scale_init_value
        self.noskip = noskip

        norm_layer = RMSNormAct2d
        self.ffn = UniversalInvertedResidual(
            in_chs=self.in_channels,
            out_chs=self.out_channels,
            dw_kernel_size_mid=0,
            exp_ratio=self.expansion_ratio,
            norm_layer=norm_layer,
            noskip=self.noskip,
            layer_scale_init_value=(
                self.layer_scale_init_value if self.use_layer_scale else None
            ),
        )

        self.norm = norm_layer(self.out_channels, eps=1e-6, apply_act=False)

    def __call__(self, inputs: list[mx.array]) -> mx.array:
        inputs = [i.transpose(0, 3, 1, 2) for i in inputs]
        high_resolution = inputs[0].shape[
            -2:
        ]  # Assuming the first input is the highest resolution.
        resized_inputs = []

        for _, img in enumerate(inputs):
            if any([r < hr for r, hr in zip(img.shape[-2:], high_resolution)]):
                img = nearest_interpolate(img, size=high_resolution)

            resized_inputs.append(img)

        channel_cat_imgs = mx.concatenate(
            resized_inputs, axis=1
        )  # Cat on channel dim, must equal self.in_channels
        img = self.ffn(channel_cat_imgs.swapaxes(1, 3)).swapaxes(1, 3)

        if any([ro != rh for ro, rh in zip(high_resolution, self.output_resolution)]):
            if (
                high_resolution[0] % self.output_resolution[0] != 0
                or high_resolution[1] % self.output_resolution[1] != 0
            ):
                img = bicubic_interpolate(img, self.output_resolution)
            else:
                h_strides = high_resolution[0] // self.output_resolution[0]
                w_strides = high_resolution[1] // self.output_resolution[1]

                img = nn.AvgPool2d(
                    kernel_size=(h_strides, w_strides),
                    stride=(h_strides, w_strides),
                )(img.swapaxes(1, 3))

            img = self.norm(img) if self.noskip else img

        return img


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/layers/layer_scale.py#L22
class LayerScale2d(nn.Module):
    def __init__(self, dim: int, init_values: float = 1e-5, inplace: bool = False):
        super().__init__()
        self.inplace = inplace
        self.gamma = init_values * mx.ones((dim,))

    def __call__(self, x: mx.array) -> mx.array:
        return x.mul_(self.gamma) if self.inplace else x * self.gamma


def rms_norm2d(
    x: mx.array,
    normalized_shape: List[int],
    weight: Optional[mx.array] = None,
    eps: float = 1e-5,
):
    assert len(normalized_shape) == 1
    dtype = x.dtype
    v = mx.power(x, 2)
    v = mx.mean(v, axis=1, keepdims=True)
    x = x * mx.rsqrt(v + eps)
    if weight is not None:
        x = x.astype(dtype) * weight.reshape(1, -1, 1, 1)
    return x


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/layers/norm_act.py#L504
class RMSNormAct2d(nn.RMSNorm):
    def __init__(
        self,
        num_channels,
        eps=1e-6,
        apply_act: bool = True,
    ):
        super().__init__(dims=num_channels, eps=eps)
        self.normalized_shape = [num_channels]
        self.drop = nn.Identity()
        self.act = nn.GELU() if apply_act else nn.Identity()

    def __call__(self, x: mx.array) -> mx.array:

        x = x.transpose(0, 3, 1, 2)  # Convert from NHWC to NCHW
        x = rms_norm2d(x, self.normalized_shape, self.weight, self.eps)
        x = self.drop(x)
        x = self.act(x)
        x = x.transpose(0, 2, 3, 1)  # Convert back to NHWC
        return x


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/models/_efficientnet_blocks.py#L310
class UniversalInvertedResidual(nn.Module):
    def __init__(
        self,
        in_chs: int,
        out_chs: int,
        dw_kernel_size_start: int = 0,
        dw_kernel_size_mid: int = 3,
        dw_kernel_size_end: int = 0,
        stride: int = 1,
        dilation: int = 1,
        group_size: int = 1,
        pad_type: str = "",
        noskip: bool = False,
        exp_ratio: float = 1.0,
        norm_layer=RMSNormAct2d,
        conv_kwargs: Optional[Dict] = None,
        drop_path_rate: float = 0.0,
        layer_scale_init_value: Optional[float] = 1e-5,
    ):
        super().__init__()
        self.has_skip = (in_chs == out_chs and stride == 1) and not noskip
        if stride > 1:
            assert dw_kernel_size_start or dw_kernel_size_mid or dw_kernel_size_end

        if dw_kernel_size_start:
            dw_start_stride = stride if not dw_kernel_size_mid else 1
            dw_start_groups = num_groups(group_size, in_chs)
            self.dw_start = ConvNormAct(
                nn.Conv2d,
                in_chs,
                in_chs,
                kernel_size=dw_kernel_size_start,
                stride=dw_start_stride,
                padding=(dw_kernel_size_start - 1) // 2,
                dilation=dilation,
                groups=dw_start_groups,
                bias=False,
                apply_act=False,
                eps=1e-05,
            )
        else:
            self.dw_start = nn.Identity()

        mid_chs = make_divisible(in_chs * exp_ratio)
        self.pw_exp = ConvNormAct(
            nn.Conv2d,
            in_chs,
            mid_chs,
            kernel_size=1,
            stride=1,
            padding=0,
            groups=1,
            bias=False,
            eps=1e-05,
        )

        if dw_kernel_size_mid:
            dw_mid_groups = num_groups(group_size, mid_chs)
            self.dw_mid = ConvNormAct(
                Conv2dSame,
                mid_chs,
                mid_chs,
                kernel_size=dw_kernel_size_mid,
                stride=stride,
                padding=0,
                dilation=dilation,
                groups=dw_mid_groups,
                bias=False,
                eps=1e-05,
            )
        else:
            self.dw_mid = nn.Identity()

        self.pw_proj = ConvNormAct(
            nn.Conv2d,
            mid_chs,
            out_chs,
            kernel_size=1,
            stride=1,
            padding=0,
            groups=1,
            bias=False,
            apply_act=False,
            eps=1e-05,
        )
        if layer_scale_init_value is not None:
            self.layer_scale = LayerScale2d(out_chs, layer_scale_init_value)
        else:
            self.layer_scale = nn.Identity()

    def __call__(self, x: mx.array) -> mx.array:
        shortcut = x
        x = self.dw_start(x)
        x = self.pw_exp(x)
        x = self.dw_mid(x)
        x = self.pw_proj(x)
        x = self.layer_scale(x)
        if self.has_skip:
            x = x + shortcut
        return x


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/layers/conv_bn_act.py#L15
class ConvNormAct(nn.Module):
    def __init__(
        self,
        conv_cls,
        in_chs: int,
        out_chs: int,
        kernel_size: int = 3,
        stride: int = 1,
        padding: int = 0,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = False,
        apply_act: bool = True,
        eps: float = 1e-6,
    ):
        super().__init__()
        self.out_chs = out_chs
        self.conv = conv_cls(
            in_chs,
            out_chs,
            kernel_size,
            stride,
            padding,
            (dilation, dilation),
            groups,
            bias,
        )
        self.bn = RMSNormAct2d(out_chs, eps=eps, apply_act=apply_act)

    def __call__(self, x: mx.array) -> mx.array:
        c = self.conv(x)
        r = self.bn(c)
        return r


def pad_same(
    x,
    kernel_size: List[int],
    stride: List[int],
    dilation: List[int] = (1, 1),
    value: float = 0,
):
    """
    Input should be in MLX format
    """
    ih, iw = x.shape[1:3]
    pad_h = get_same_padding(ih, kernel_size[0], stride[0], dilation[0])
    pad_w = get_same_padding(iw, kernel_size[1], stride[1], dilation[1])

    # MLX pad format: [(low, high), (low, high), ...] for each axis
    # Padding order is reversed compared to PyTorch F.pad
    pad_widths = [
        (0, 0),  # No padding for batch dimension
        (pad_h // 2, pad_h - pad_h // 2),  # Height padding
        (pad_w // 2, pad_w - pad_w // 2),  # Width padding
        (0, 0),  # No padding for channel dimension
    ]

    x = mx.pad(x, pad_widths, constant_values=value)
    return x


def get_padding_value(padding, kernel_size, **kwargs) -> Tuple[Tuple, bool]:
    dynamic = False
    if isinstance(padding, str):
        # for any string padding, the padding will be calculated for you, one of three ways
        padding = padding.lower()
        if padding == "same":
            # TF compatible 'SAME' padding, has a performance and GPU memory allocation impact
            if is_static_pad(kernel_size, **kwargs):
                # static case, no extra overhead
                padding = get_padding(kernel_size, **kwargs)
            else:
                # dynamic 'SAME' padding, has runtime/GPU memory overhead
                padding = 0
                dynamic = True
        elif padding == "valid":
            # 'VALID' padding, same as padding=0
            padding = 0
        else:
            # Default to PyTorch style 'same'-ish symmetric padding
            padding = get_padding(kernel_size, **kwargs)
    return padding, dynamic


def get_same_padding(
    input_size: int, kernel_size: int, stride: int, dilation: int = 1
) -> int:
    """Calculate padding needed for 'same' output size."""
    effective_kernel_size = dilation * (kernel_size - 1) + 1
    output_size = (input_size + stride - 1) // stride
    total_padding = max(
        0, (output_size - 1) * stride + effective_kernel_size - input_size
    )
    return total_padding


def get_padding(kernel_size, stride=1, dilation=1, **_):
    """Get symmetric padding for given kernel size."""
    if isinstance(kernel_size, int):
        kernel_size = [kernel_size, kernel_size]
    if isinstance(stride, int):
        stride = [stride, stride]
    if isinstance(dilation, int):
        dilation = [dilation, dilation]

    padding = []
    for k, d in zip(kernel_size, dilation):
        effective_k = d * (k - 1) + 1
        pad_total = effective_k - 1
        padding.append(pad_total // 2)
    return tuple(padding)


def is_static_pad(kernel_size, stride=1, dilation=1, **_):
    """Check if padding can be calculated statically."""
    if isinstance(kernel_size, int):
        kernel_size = [kernel_size, kernel_size]
    if isinstance(stride, int):
        stride = [stride, stride]
    if isinstance(dilation, int):
        dilation = [dilation, dilation]

    # Static padding is possible when stride is 1 for all dimensions
    return all(s == 1 for s in stride)


class Conv2dSame(nn.Conv2d):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kernel_size = self.weight.shape[1:3]

    def __call__(self, x: mx.array) -> mx.array:
        x = pad_same(x, self.kernel_size, self.stride, self.dilation)
        y = mx.conv2d(
            x, self.weight, self.stride, self.padding, self.dilation, self.groups
        )
        if "bias" in self:
            y = y + self.bias
        return y


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/models/_efficientnet_blocks.py#L629
class EdgeResidual(nn.Module):
    def __init__(
        self,
        in_chs: int,
        out_chs: int,
        exp_kernel_size: int = 3,
        stride: int = 1,
        dilation: int = 1,
        group_size: int = 0,
        pad_type: str = "",
        force_in_chs: int = 0,
        noskip: bool = False,
        expand_ratio: float = 1.0,
        pw_kernel_size: int = 1,
        norm_layer=RMSNormAct2d,
    ):
        super().__init__()

        if force_in_chs > 0:
            mid_chs = make_divisible(force_in_chs * expand_ratio)
        else:
            mid_chs = make_divisible(in_chs * expand_ratio)

        groups = num_groups(group_size, mid_chs)

        self.has_skip = (in_chs == out_chs and stride == 1) and not noskip

        self.conv_exp = Conv2dSame(
            in_chs,
            mid_chs,
            kernel_size=exp_kernel_size,
            stride=stride,
            padding=0,
            dilation=(dilation, dilation),
            groups=groups,
            bias=False,
        )

        self.bn1 = norm_layer(mid_chs, eps=1e-05) if norm_layer else nn.Identity()

        # Point-wise linear projection
        padding_pwl = (pw_kernel_size - 1) // 2
        self.conv_pwl = nn.Conv2d(
            mid_chs,
            out_chs,
            kernel_size=pw_kernel_size,
            padding=padding_pwl,
            bias=False,
        )

        self.bn2 = (
            norm_layer(out_chs, eps=1e-05, apply_act=False)
            if norm_layer
            else nn.Identity()
        )

    def __call__(self, x: mx.array) -> mx.array:
        shortcut = x
        x = self.conv_exp(x)
        x = self.bn1(x)
        x = self.conv_pwl(x)
        x = self.bn2(x)
        if self.has_skip:
            x = x + shortcut
        return x


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/models/_efficientnet_blocks.py#L449
class MobileAttention(nn.Module):
    def __init__(
        self,
        in_chs: int,
        out_chs: int,
        stride: int = 1,
        dw_kernel_size: int = 3,
        dilation: int = 1,
        group_size: int = 1,
        pad_type: str = "",
        num_heads: int = 8,
        key_dim: int = 64,
        value_dim: int = 64,
        use_multi_query: bool = True,
        query_strides: Tuple[int, int] = (1, 1),
        kv_stride: int = 1,
        cpe_dw_kernel_size: int = 3,
        noskip: bool = False,
        act_layer=nn.GELU,
        aa_layer=None,
        drop_path_rate: float = 0.0,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        layer_scale_init_value: Optional[float] = 1e-5,
        use_bias: bool = False,
    ):
        super().__init__()
        self.has_skip = (stride == 1 and in_chs == out_chs) and not noskip
        self.query_strides = to_2tuple(query_strides)
        self.kv_stride = kv_stride
        self.has_query_stride = any([s > 1 for s in self.query_strides])

        # Normalization layer
        self.norm = RMSNormAct2d(
            in_chs,
            eps=1e-05,
            apply_act=False,
        )
        # Determine number of heads if not provided
        if num_heads is None:
            assert in_chs % key_dim == 0
            num_heads = in_chs // key_dim

        # Attention layer
        if use_multi_query:
            self.attn = MultiQueryAttention2d(
                in_chs,
                dim_out=out_chs,
                num_heads=num_heads,
                key_dim=key_dim,
                value_dim=value_dim,
                query_strides=query_strides,
                kv_stride=kv_stride,
                dilation=dilation,
                padding=pad_type,
                dw_kernel_size=dw_kernel_size,
                attn_drop=attn_drop,
                proj_drop=proj_drop,
            )
        else:
            raise NotImplementedError("attention not implemented")

        # Layer scaling
        if layer_scale_init_value is not None:
            self.layer_scale = LayerScale2d(out_chs, layer_scale_init_value)
        else:
            self.layer_scale = nn.Identity()

        # Drop path for residual connection
        self.drop_path = nn.Identity()

    def __call__(self, x: mx.array) -> mx.array:
        shortcut = x
        x = self.norm(x)
        x = self.attn(x)
        x = self.layer_scale(x)

        # Apply skip connection if available
        if self.has_skip:
            x = self.drop_path(x) + shortcut
        return x


def create_conv2d(
    in_channels,
    out_channels,
    kernel_size,
    stride=1,
    dilation=1,
    depthwise=False,
    bias=False,
    **kwargs,
):
    """Helper function to create a 2D convolution with common parameters"""
    if depthwise:
        # Depthwise convolution
        return nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2 * dilation,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
        )
    else:
        # Regular convolution
        return nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=(kernel_size - 1) // 2 * dilation,
            dilation=dilation,
            bias=bias,
        )


def to_2tuple(x):
    """Convert input to 2-tuple"""
    if isinstance(x, tuple):
        return x
    return (x, x)


class NamedSequential(nn.Module):
    def __init__(self):
        super().__init__()
        self._order = []

    def add_module(self, name, module):
        setattr(self, name, module)
        self._order.append(name)

    def __call__(self, x):
        for name in self._order:
            x = getattr(self, name)(x)
        return x


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/layers/attention2d.py#L82
class MultiQueryAttention2d(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: Optional[int] = None,
        num_heads: int = 8,
        key_dim: int = 64,
        value_dim: int = 64,
        query_strides: Tuple[int, int] = (1, 1),
        kv_stride: int = 1,
        dilation: int = 1,
        padding: str = "",
        dw_kernel_size: int = 3,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
    ):
        super().__init__()
        dim_out = dim_out or dim
        self.num_heads = num_heads
        self.query_strides = to_2tuple(query_strides)
        self.kv_stride = kv_stride
        self.fused_attn = True
        self.key_dim = key_dim
        self.value_dim = value_dim
        head_dim = key_dim
        self.scale = head_dim**-0.5

        self.query = NamedSequential()
        self.query.add_module(
            "proj",
            create_conv2d(
                dim,
                self.num_heads * self.key_dim,
                kernel_size=1,
            ),
        )
        self.key = NamedSequential()
        if kv_stride > 1:
            self.key.add_module(
                "down_conv",
                create_conv2d(
                    dim,
                    dim,
                    kernel_size=dw_kernel_size,
                    stride=kv_stride,
                    dilation=dilation,
                    padding=padding,
                    depthwise=True,
                ),
            )
            self.key.add_module("norm", RMSNormAct2d(dim, eps=1e-6, apply_act=False))
        self.key.add_module(
            "proj", create_conv2d(dim, key_dim, kernel_size=1, bias=False)
        )

        self.value = NamedSequential()
        if kv_stride > 1:
            self.value.add_module(
                "down_conv",
                create_conv2d(
                    dim,
                    dim,
                    kernel_size=dw_kernel_size,
                    stride=kv_stride,
                    dilation=dilation,
                    padding=padding,
                    depthwise=True,
                ),
            )
            self.value.add_module("norm", RMSNormAct2d(dim, eps=1e-6, apply_act=False))
        self.value.add_module(
            "proj", create_conv2d(dim, value_dim, kernel_size=1, bias=False)
        )

        # Attention dropout
        self.attn_drop = nn.Dropout(attn_drop) if attn_drop > 0 else nn.Identity()

        # Output projection
        self.output = NamedSequential()
        self.output.add_module(
            "proj",
            create_conv2d(
                value_dim * num_heads,
                dim_out,
                kernel_size=1,
                stride=1,
                bias=False,
            ),
        )
        self.proj_drop = nn.Dropout(proj_drop) if proj_drop > 0 else nn.Identity()

    def _reshape_input(self, t: mx.array):
        """
        Input shape MLX: [B, H, W, C]
        Input shape PyTorch: [B, C, H, W]

        PyTorch Reshape: [B, C, H, W] -> [B, C, -1] -> [B, -1, C] -> [B, 1, -1, C] -> SDPA
        MLX Reshape: [B, H, W, C] -> [B, -1, C] -> [B, 1, -1, C] -> SDPA
        """
        s = t.shape
        t = t.reshape(s[0], -1, s[3])[:, None, :, :]

        return t

    def _reshape_projected_query(self, t: mx.array, num_heads: int, key_dim: int):
        """
        Input shape MLX: [B, H, W, C] where C = num_heads * key_dim
        """
        B, H, W, C = t.shape
        # t = t.reshape(B, H, W, num_heads, key_dim)
        t = t.reshape(B, H * W, num_heads, key_dim)
        return t.transpose(0, 2, 1, 3)

    def _reshape_output(self, t: mx.array, num_heads: int, h_px: int, w_px: int):
        """
        Input shape: [B, NH, L, D] where L = h_px * w_px
        Output shape MLX: [B, H, W, C] where C = NH * D
        """
        B, NH, L, D = t.shape
        # First transpose to [B, L, NH, D]
        t = t.transpose(0, 2, 1, 3)
        # Then reshape to [B, H, W, NH*D]
        t = t.reshape(B, h_px, w_px, NH * D)
        return t

    def __call__(self, x: mx.array, attn_mask: Optional[mx.array] = None) -> mx.array:
        B, H, W, C = x.shape
        q = self.query(x)
        q = self._reshape_projected_query(q, self.num_heads, self.key_dim)

        k = self.key(x)
        k = self._reshape_input(k)

        v = self.value(x)
        v = self._reshape_input(v)

        if self.fused_attn:
            o = mx.fast.scaled_dot_product_attention(
                q,
                k,
                v,
                scale=1.0 / sqrt(q.shape[-1]),
            )
        else:
            raise NotImplementedError("unfused attention not implemented")

        o = self._reshape_output(
            o, self.num_heads, H // self.query_strides[0], W // self.query_strides[1]
        )
        x = self.output(o)
        return x


def num_groups(group_size: Optional[int], channels: int) -> int:
    if not group_size:  # 0 or None
        return 1  # normal conv with 1 group
    else:
        # NOTE group_size == 1 -> depthwise conv
        assert channels % group_size == 0
        return channels // group_size


def make_divisible(v, divisor: int = 8, min_value=None, round_limit: float = 0.9):
    min_value = min_value or divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < round_limit * v:
        new_v += divisor
    return new_v


def _er(kernel_size, filters, strides=1, expand_ratio=4.0, is_multiscale=False):
    return EdgeResidualConfig(
        kernel_size=kernel_size,
        filters=filters,
        strides=strides,
        expand_ratio=expand_ratio,
        is_multiscale=is_multiscale,
    )


def _uir(
    start_dw_kernel_size,
    mid_dw_kernel_size,
    filters,
    strides=1,
    expand_ratio=4.0,
    is_multiscale=False,
):
    return UniversalInvertedResidualConfig(
        start_dw_kernel_size=start_dw_kernel_size,
        mid_dw_kernel_size=mid_dw_kernel_size,
        filters=filters,
        strides=strides,
        expand_ratio=expand_ratio,
        is_multiscale=is_multiscale,
    )


def _mmqa(
    num_heads,
    kv_dim,
    kv_strides,
    mmqa_avg_pool_kv=False,
    is_multiscale=False,
):
    conf = MultiQueryAttentionBlockConfig(
        num_heads=num_heads,
        kv_dim=kv_dim,
        kv_strides=kv_strides,
        mmqa_avg_pool_kv=mmqa_avg_pool_kv,
        is_multiscale=is_multiscale,
    )
    return conf


# https://github.com/huggingface/new-model-addition-timm-gemma3p5-non-fork/blob/mobilenet-gemma3n-rw/timm/models/mobilenetv5.py#L596
def gemma3n_mobilenet_def():
    return [
        # Stage 1: Edge Residuals
        [_er(3, 128, 2)] + [_er(3, 128, 1)] * 2,
        # Stage 2: Universal Inverted Residuals
        [_uir(3, 5, 256, 2, 6.0)] + [_uir(k, 0, 256) for k in [5, 3, 5, 3]],
        # Stage 3: Universal Inverted Residuals with Multi-Query Attention
        [_uir(5, 5, 640, 2, 6.0)]
        + [_uir(5, 0, 640)] * 7
        + [_uir(0, 0, 640, 1, 1.0)]
        + [_mmqa(12, 64, 2), _uir(0, 0, 640, 1, 2.0)] * 13
        + [_mmqa(12, 64, 2), _uir(0, 0, 640, 1, 2.0, is_multiscale=True)],
        # Stage 4: Universal Inverted Residuals with Multi-Query Attention
        [_uir(5, 5, 1280, 2, 6.0)]
        + [_mmqa(16, 96, 1), _uir(0, 0, 1280, 1, 2.0)] * 18
        + [_mmqa(16, 96, 1), _uir(0, 0, 1280, 1, 2.0, is_multiscale=True)],
    ]


class VisionTower(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.conv_stem = ConvNormAct(
            Conv2dSame,
            in_chs=3,
            out_chs=64,
            kernel_size=3,
            stride=2,
            padding=0,
            eps=1e-05,
            bias=True,
        )
        msfa_indices = (3, 4)
        msfa_output_resolution = (16, 16)

        (num_features, self.blocks) = self.build()
        self.num_features = self.head_hidden_size = (
            num_features  # output of msfa is output of forward_features()
        )
        self.msfa_indices = msfa_indices
        self.msfa_output_resolution = msfa_output_resolution

        self.msfa = MobileNetV5MultiScaleFusionAdapter(
            in_chs=[1920],
            out_chs=2048,
            output_resolution=self.msfa_output_resolution,
        )

    def build(self):
        blocks = []
        in_chs = self.conv_stem.out_chs
        for stage, block_config in enumerate(gemma3n_mobilenet_def()):
            block_group = []
            for config in block_config:
                match config:
                    case EdgeResidualConfig(
                        kernel_size, filters, strides, expand_ratio, is_multiscale
                    ):
                        x = EdgeResidual(
                            exp_kernel_size=kernel_size,
                            in_chs=in_chs,
                            out_chs=filters,
                            stride=strides,
                            expand_ratio=expand_ratio,
                        )
                        in_chs = filters  # in_chs of next is out_chs of prev
                        block_group.append(x)
                    case UniversalInvertedResidualConfig(
                        start_dw_kernel_size,
                        mid_dw_kernel_size,
                        filters,
                        strides,
                        expand_ratio,
                        is_multiscale,
                    ):
                        x = UniversalInvertedResidual(
                            in_chs=in_chs,
                            out_chs=filters,
                            dw_kernel_size_start=start_dw_kernel_size,
                            dw_kernel_size_mid=mid_dw_kernel_size,
                            stride=strides,
                            exp_ratio=expand_ratio,
                        )
                        in_chs = filters
                        block_group.append(x)
                    case MultiQueryAttentionBlockConfig(
                        num_heads,
                        kv_dim,
                        kv_strides,
                        mmqa_avg_pool_kv,
                        is_multiscale,
                    ):
                        x = MobileAttention(
                            in_chs=in_chs,
                            out_chs=in_chs,
                            stride=1,
                            num_heads=num_heads,
                            key_dim=kv_dim,
                            value_dim=kv_dim,
                            kv_stride=kv_strides,
                            act_layer=None,
                        )
                        block_group.append(x)
                    case _:
                        continue
            blocks.append(block_group)
        return (in_chs, blocks)

    def __call__(
        self, x: mx.array, output_hidden_states: Optional[bool] = None
    ) -> mx.array:
        feat_idx = 0
        x = x.transpose(0, 2, 3, 1)  # Convert from NCHW to NHWC
        x = self.conv_stem(x)
        intermediates = []

        if feat_idx in self.msfa_indices:
            intermediates.append(x)

        # MBV5 is constructed of 4 stages, each stage is a group of blocks.
        for block_group in self.blocks:
            feat_idx += 1
            for block in block_group:
                x = block(x)

            if feat_idx in self.msfa_indices:
                intermediates.append(x)

        x = self.msfa(intermediates)
        return x


class VisionModel(nn.Module):
    def __init__(self, config: VisionConfig):
        super().__init__()
        self.model_type = config.model_type
        if self.model_type not in ["gemma3", "gemma3_vision", "gemma3n_vision"]:
            raise ValueError(f"Unsupported model type: {self.model_type}")

        self.timm_model = VisionTower(config)

    def __call__(
        self, x: mx.array, output_hidden_states: Optional[bool] = None
    ) -> mx.array:
        return self.timm_model(x, output_hidden_states)

    def sanitize(self, weights):
        sanitized_weights = {}
        skip_transpose = False
        _, H, _, C = weights["vision_tower.timm_model.blocks.0.0.conv_exp.weight"].shape
        if C > H:
            skip_transpose = True

        for k, v in weights.items():
            # PyTorch conv2d weight: [out_channels, in_channels, kH, kW]
            # MLX conv2d weight: [out_channels, kH, KW, in_channels]
            if ("conv" in k and "weight" in k) or ("attn" and "proj.weight") in k:
                if len(v.shape) == 4 and not skip_transpose:
                    v = v.transpose(0, 2, 3, 1)
            sanitized_weights[k] = v

        return sanitized_weights



import argparse
import os
import sys
import time
from typing import Dict, List

import mlx.core as mx
from rich import print as rprint
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.prompt import Prompt

from mlx_vlm import load
from mlx_vlm.generate import generate_step
from mlx_vlm.prompt_utils import get_message_json
from mlx_vlm.utils import load_image


class MLXVisionChat:
    def __init__(
        self,
        model_path: str = "mlx-community/idefics2-8b-chatty-4bit",
        temperature: float = 0.7,
        max_tokens: int = 1000,
        verbose: bool = False,
    ):
        self.console = Console()
        self.verbose = verbose
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.history: List[Dict] = []
        self.current_image = None

        with self.console.status("[bold green]Loading model..."):
            self.model, self.processor = load(model_path)

        rprint("[bold green]Model loaded successfully![/bold green]")
        self.print_help()

    def print_help(self) -> None:
        """Print available commands."""
        help_text = """
[bold yellow]Available Commands:[/bold yellow]
• /image <path> - Load a new image for discussion
• /clear - Clear conversation history
• /help - Show this help message
• /exit - Exit the chat
• Any other input will be treated as a question or comment about the current image
        """
        rprint(Panel(help_text, title="Help", border_style="blue"))

    def process_image(self, image_path: str) -> bool:
        """Process an image and prepare it for the model. Returns True if successful."""
        try:
            if not os.path.exists(image_path):
                rprint(
                    f"[bold red]Error:[/bold red] Image file not found: {image_path}"
                )
                return False

            self.current_image = load_image(image_path)
            rprint(f"[bold blue]Loaded image:[/bold blue] {image_path}")
            return True
        except Exception as e:
            rprint(f"[bold red]Error loading image:[/bold red] {str(e)}")
            return False

    def add_to_history(self, role: str, text: str) -> None:
        """Add a message to the conversation history."""
        content = [{"type": "text", "text": text}]
        self.history.append({"role": role, "content": content})

    def generate_response(self) -> str:
        """Generate a response from the model based on the conversation history."""
        if self.current_image is None:
            return "Please load an image first using the /image command."

        messages = []
        for i, message in enumerate(self.history):
            skip_token = True
            if i == len(self.history) - 1 and message["role"] == "user":
                skip_token = False
            messages.append(
                get_message_json(
                    self.model.config.model_type,
                    message["content"][0]["text"],
                    role=message["role"],
                    skip_image_token=skip_token,
                )
            )

        text_prompt = self.processor.apply_chat_template(
            messages, add_generation_prompt=True
        )

        inputs = self.processor(
            text=[text_prompt],
            images=[self.current_image],
            padding=True,
            return_tensors="np",
        )

        pixel_values = mx.array(inputs["pixel_values"])
        input_ids = mx.array(inputs["input_ids"])
        mask = mx.array(inputs["attention_mask"])

        detokenizer = self.processor.detokenizer
        detokenizer.reset()

        tic = time.perf_counter()

        generator = generate_step(
            input_ids,
            self.model,
            pixel_values,
            mask,
            temperature=self.temperature,
        )

        # Use print instead of rprint to avoid rich console's automatic newlines
        rprint("[bold green]Assistant:[/bold green]", end=" ", flush=True)
        for (token, prob), n in zip(generator, range(self.max_tokens)):
            if n == 0:
                prompt_time = time.perf_counter() - tic
                tic = time.perf_counter()

            if token == self.processor.tokenizer.eos_token_id and n > 0:
                break

            detokenizer.add_token(token)

            if self.verbose:
                rprint(detokenizer.last_segment, end="", flush=True)

        detokenizer.finalize()
        return detokenizer.text

    def handle_command(self, command: str, args: str) -> bool:
        """Handle special commands. Returns True if should continue chat, False if should exit."""
        if command == "/exit":
            rprint("[bold yellow]Goodbye![/bold yellow]")
            return False
        elif command == "/help":
            self.print_help()
        elif command == "/clear":
            self.history.clear()
            rprint("[bold blue]Conversation history cleared.[/bold blue]")
        elif command == "/image":
            if not args:
                rprint("[bold red]Error:[/bold red] Please provide an image path")
                return True
            self.process_image(args.strip())
        else:
            rprint(f"[bold red]Unknown command:[/bold red] {command}")
        return True

    def chat_loop(self) -> None:
        """Main chat loop for interaction."""
        while True:
            try:
                user_input = Prompt.ask("\n[bold cyan]You[/bold cyan]").strip()

                # Handle commands
                if user_input.startswith("/"):
                    parts = user_input.split(maxsplit=1)
                    command = parts[0].lower()
                    args = parts[1] if len(parts) > 1 else ""
                    if not self.handle_command(command, args):
                        break
                    continue
                # Handle regular chat input
                if self.current_image is None:
                    rprint(
                        "[bold yellow]Please load an image first using the /image command[/bold yellow]"
                    )
                    continue

                self.add_to_history("user", user_input)
                response = self.generate_response()

                if not self.verbose:
                    rprint(Panel(Markdown(response), border_style="green"))

                # Remove the eos token from the response
                response = response.replace("<end_of_utterance>", "")

                self.add_to_history("assistant", response)

            except KeyboardInterrupt:
                rprint(
                    "\n[bold yellow]Interrupted by user. Type /exit to quit.[/bold yellow]"
                )
                continue
            except Exception as e:
                rprint(f"[bold red]Error:[/bold red] {str(e)}")
                continue


def main():
    parser = argparse.ArgumentParser(description="MLX Vision Chat CLI")
    parser.add_argument(
        "--model",
        default="mlx-community/idefics2-8b-chatty-4bit",
        help="Path to the model or model identifier",
    )
    parser.add_argument("--verbose", action="store_false", help="Enable verbose output")
    parser.add_argument(
        "--temperature", type=float, default=0.7, help="Temperature for the model"
    )
    parser.add_argument(
        "--max-tokens",
        type=int,
        default=1000,
        help="Maximum number of new tokens to generate",
    )

    args = parser.parse_args()

    try:
        chat = MLXVisionChat(
            model_path=args.model,
            temperature=args.temperature,
            max_tokens=args.max_tokens,
            verbose=args.verbose,
        )
        chat.chat_loop()
    except Exception as e:
        rprint(f"[bold red]Fatal error:[/bold red] {str(e)}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import mlx.core as mx
import numpy as np

from ..idefics3 import LanguageModel
from ..idefics3 import Model as Idefics3Model
from ..idefics3 import VisionModel
from .config import ModelConfig, TextConfig, VisionConfig


class Model(Idefics3Model):
    def _prepare_inputs_for_multimodal(self, image_features, inputs_embeds, input_ids):
        # Assumes bs == 1

        B, T, D_text = inputs_embeds.shape
        N, S, D_img = image_features.shape

        image_offset = 0
        cur_embeds = inputs_embeds[0]

        # Find positions of <image> tokens in the text
        image_token_index = self.config.image_token_index
        image_positions = np.where(input_ids == image_token_index)[1].tolist()
        num_image_tokens = len(image_positions)

        # If no <image> => text-only
        if num_image_tokens == 0:
            empty_slice = image_features[0][:0, :]  # shape (0, D)
            return mx.concatenate([cur_embeds, empty_slice], axis=0)

        # Typically, if each image is S embeddings, we expect the total # of <image> tokens
        # in this sample to be multiple of S => each group of S tokens = 1 image
        if num_image_tokens % S != 0:
            raise ValueError(
                f"Input has {num_image_tokens} <image> tokens, not a multiple of S={S}. "
                "Cannot map them to blocks of shape (S, D)."
            )

        chunks = [image_positions[i : i + S] for i in range(0, num_image_tokens, S)]

        segments = []
        text_start = 0

        # For each chunk (each chunk => 1 image)
        for chunk in chunks:
            cur_block = image_features[image_offset]
            image_offset += 1

            # We'll iterate over the S positions in ascending order
            for i_s, pos in enumerate(chunk):
                if pos > text_start:
                    segments.append(cur_embeds[text_start:pos])
                # Then add one row from cur_block => shape (1, D)
                row_of_block = cur_block[i_s : i_s + 1, :]
                segments.append(row_of_block)
                text_start = pos + 1

        # leftover text after the final <image> token
        if text_start < T:
            segments.append(cur_embeds[text_start:])

        # cat them into a single (T_b, D) tensor
        merged_sample = mx.concatenate(segments, axis=0)
        return mx.expand_dims(merged_sample, axis=0)



import glob
import inspect
import json
import math
from pathlib import Path
from typing import List, Optional, Tuple, Union

import mlx.core as mx
import mlx.nn as nn
from huggingface_hub import snapshot_download
from mlx.utils import tree_map

from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


def shift_tokens_right(
    input_ids: mx.array, pad_token_id: int, decoder_start_token_id: int
) -> mx.array:
    """Shift input tokens right, adding decoder start token at beginning."""
    shifted = mx.roll(input_ids, 1, axis=-1)
    shifted = tree_map(lambda x: x.at[:, 0].set(decoder_start_token_id), shifted)
    shifted = mx.where(shifted == -100, pad_token_id, shifted)
    return shifted


class LearnedPositionEmbedding2D(nn.Module):
    """2D learned position embeddings."""

    def __init__(self, embedding_dim: int = 256, num_pos: int = 50):
        super().__init__()
        self.row_embeddings = nn.Embedding(num_pos, embedding_dim // 2)
        self.column_embeddings = nn.Embedding(
            num_pos, embedding_dim - (embedding_dim // 2)
        )

    def __call__(self, x):
        batch_size, height, width, channels = x.shape
        width_pos = mx.arange(width)
        height_pos = mx.arange(height)

        x_emb = self.column_embeddings(width_pos)
        y_emb = self.row_embeddings(height_pos)

        pos = mx.concatenate(
            [
                mx.broadcast_to(x_emb[None, :, :], (height, width, x_emb.shape[-1])),
                mx.broadcast_to(y_emb[:, None, :], (height, width, y_emb.shape[-1])),
            ],
            axis=-1,
        )

        return mx.broadcast_to(pos[None, ...], (batch_size, height, width, channels))


class PositionalEmbeddingCosine1D(nn.Module):
    """
    MLX implementation of 1D cosine positional embeddings.

    Args:
        embed_dim: The dimension of the embeddings
        max_seq_len: The maximum length to precompute the positional encodings
    """

    def __init__(self, embed_dim: int = 512, max_seq_len: int = 1024) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.max_seq_len = max_seq_len

        # Generate position indices and dimension indices
        position = mx.arange(max_seq_len)
        dim_pos = mx.arange(0, embed_dim // 2)  # Half the dimensions for sin/cos pairs

        # Calculate frequency bands
        factor = math.log(10000)
        denominator = mx.exp(-factor * dim_pos / embed_dim)

        # Create position-frequency product matrix [max_seq_len, embed_dim//2]
        frequencies = mx.reshape(position, (-1, 1)) * denominator

        # Calculate sin and cos values [max_seq_len, embed_dim//2]
        sin_values = mx.sin(frequencies)
        cos_values = mx.cos(frequencies)

        # Interleave sin and cos values to create final embeddings
        pos_idx_to_embed = mx.zeros((max_seq_len, embed_dim))
        pos_idx_to_embed = mx.concatenate(
            [mx.expand_dims(sin_values, -1), mx.expand_dims(cos_values, -1)], axis=-1
        ).reshape(max_seq_len, embed_dim)

        # Store the positional embeddings
        self.pos_idx_to_embed = pos_idx_to_embed

    def __call__(self, seq_embeds: mx.array) -> mx.array:
        """
        Apply positional embeddings to the input sequence.

        Args:
            seq_embeds: Input sequence embeddings with shape:
                - [T, D] where T is sequence length and D is embedding dimension
                - [B, T, D] where B is batch size

        Returns:
            Positional embeddings matching input shape
        """
        shape_len = len(seq_embeds.shape)
        assert 2 <= shape_len <= 3, "Input must be 2D or 3D tensor"

        len_seq = seq_embeds.shape[-2]
        assert (
            len_seq <= self.max_seq_len
        ), f"Sequence length {len_seq} exceeds maximum length {self.max_seq_len}"

        # Get relevant portion of pre-computed embeddings
        pos_embeds = self.pos_idx_to_embed[:len_seq]

        # Add batch dimension if input is 3D
        if shape_len == 3:
            pos_embeds = mx.expand_dims(pos_embeds, 0)

        return pos_embeds


class Model(nn.Module):
    """Florence-2 model for conditional generation."""

    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config

        # Initialize vision model
        self.vision_tower = VisionModel(config.vision_config)

        # Initialize language model
        self.language_model = LanguageModel(config.text_config)

        # Image projection layers
        image_dim = config.vision_config.dim_embed[-1]
        text_dim = config.text_config.d_model
        self.image_projection = mx.zeros((image_dim, text_dim))

        self.image_proj_norm = nn.LayerNorm(text_dim)

        # Position embeddings
        if config.image_pos_embed["type"] == "learned_abs_2d":
            self.image_pos_embed = LearnedPositionEmbedding2D(
                embedding_dim=image_dim,
                num_pos=config.image_pos_embed["max_pos_embeddings"],
            )
        else:
            raise NotImplementedError(
                f"Position embedding type {config.image_pos_embed['type']} not supported"
            )

        # Temporal embeddings
        if config.visual_temporal_embedding["type"] == "COSINE":
            self.visual_temporal_embed = PositionalEmbeddingCosine1D(
                embed_dim=image_dim,
                max_seq_len=config.visual_temporal_embedding["max_temporal_embeddings"],
            )
        else:
            raise NotImplementedError(
                f"Temporal embedding type {config.visual_temporal_embedding['type']} not supported"
            )

        self.image_feature_source = config.image_feature_source

    def _encode_image(self, pixel_values, extract_features=True):
        """Encode image using vision model and add position embeddings."""
        T = 1  # Single frame for now

        # Get vision features
        if extract_features:
            batch_size, C, H, W = pixel_values.shape
            x = self.vision_tower(pixel_values)
        else:
            x = pixel_values
            batch_size = pixel_values.shape[0]

        # Assuming this is part of a class method, keeping the same structure
        if self.image_pos_embed is not None:
            # Reshape to (batch_size * T, -1, feature_dim)
            x = mx.reshape(x, (batch_size * T, -1, x.shape[-1]))
            num_tokens = x.shape[-2]
            h, w = int(num_tokens**0.5), int(num_tokens**0.5)
            assert h * w == num_tokens, "only support square feature maps for now"
            # Reshape to (batch_size * T, h, w, feature_dim)
            x = mx.reshape(x, (batch_size * T, h, w, x.shape[-1]))
            pos_embed = self.image_pos_embed(x)
            x = x + pos_embed
            # Reshape to (batch_size, T * h * w, feature_dim)
            x = mx.reshape(x, (batch_size, T * h * w, x.shape[-1]))

        if self.visual_temporal_embed is not None:
            # Reshape for temporal embedding
            x_temp = mx.reshape(x, (batch_size, T, -1, x.shape[-1]))
            temporal_input = x_temp[:, :, 0]
            visual_temporal_embed = self.visual_temporal_embed(temporal_input)
            # Expand dims for broadcasting
            visual_temporal_embed = mx.expand_dims(visual_temporal_embed, axis=2)
            x = mx.reshape(x, (batch_size, T, -1, x.shape[-1])) + visual_temporal_embed

        x_feat_dict = {}

        # Spatial average pooling
        x_spatial = mx.reshape(x, (batch_size, T, -1, x.shape[-1]))
        spatial_avg_pool_x = mx.mean(x_spatial, axis=2)
        x_feat_dict["spatial_avg_pool"] = spatial_avg_pool_x

        # Temporal average pooling
        x_temporal = mx.reshape(x, (batch_size, T, -1, x.shape[-1]))
        temporal_avg_pool_x = mx.mean(x_temporal, axis=1)
        x_feat_dict["temporal_avg_pool"] = temporal_avg_pool_x

        # Last frame features
        x_last = mx.reshape(x, (batch_size, T, -1, x.shape[-1]))
        x = x_last[:, -1]
        x_feat_dict["last_frame"] = x

        # Gather features based on source configuration
        new_x = []
        for _image_feature_source in self.image_feature_source:
            if _image_feature_source not in x_feat_dict:
                raise ValueError(
                    f"invalid image feature source: {_image_feature_source}"
                )
            new_x.append(x_feat_dict[_image_feature_source])

        # Concatenate features
        x = mx.concatenate(new_x, axis=1)

        # Final projection and normalization
        x = x @ self.image_projection
        x = self.image_proj_norm(x)

        return x

    def _merge_input_ids_with_image_features(self, image_features, inputs_embeds=None):
        batch_size, image_token_length, _ = image_features.shape
        image_attention_mask = mx.ones((batch_size, image_token_length))

        if inputs_embeds is None:
            return image_features, image_attention_mask

        task_prefix_embeds = inputs_embeds
        task_prefix_attention_mask = mx.ones((batch_size, task_prefix_embeds.shape[1]))

        if len(task_prefix_attention_mask.shape) == 3:
            task_prefix_attention_mask = task_prefix_attention_mask[:, 0]

        # Concatenate image features and task prefix embeddings
        inputs_embeds = mx.concatenate([image_features, task_prefix_embeds], axis=1)
        attention_mask = mx.concatenate(
            [image_attention_mask, task_prefix_attention_mask], axis=1
        )
        return inputs_embeds, attention_mask

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids=None,
        pixel_values=None,
        cache=None,
        decoder_input_ids=None,
        decoder_attention_mask=None,
        labels=None,
        **kwargs,
    ):
        """Forward pass."""
        attention_mask = None
        decoder_inputs_embeds = None

        # Process image if provided
        if pixel_values is not None:
            image_features = self._encode_image(pixel_values)

            # Get input embeddings if needed
            inputs_embeds = None
            if input_ids is not None:
                inputs_embeds = self.language_model.model.shared(input_ids)

            # Merge image features with text embeddings
            inputs_embeds, attention_mask = self._merge_input_ids_with_image_features(
                image_features, inputs_embeds
            )
        else:
            inputs_embeds = None
            attention_mask = None

        # Handle decoder input IDs
        if labels is not None and decoder_input_ids is None:
            decoder_input_ids = shift_tokens_right(
                labels, self.config.pad_token_id, self.config.bos_token_id
            )

        if decoder_input_ids is None and decoder_inputs_embeds is None:
            decoder_start_token_id = getattr(
                self.config, "decoder_start_token_id", 0
            )  # 2 is common for many models
            if decoder_start_token_id is None:
                decoder_start_token_id = 0

            decoder_input_ids = mx.array([decoder_start_token_id])[None, :]
            decoder_inputs_embeds = self.language_model.model.shared(decoder_input_ids)
            decoder_input_ids = None

        # Forward through language model
        outputs = self.language_model(
            input_ids=input_ids,
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            decoder_input_ids=decoder_input_ids,
            decoder_inputs_embeds=decoder_inputs_embeds,
            decoder_attention_mask=decoder_attention_mask,
            cache=cache,
        )

        return outputs

    @staticmethod
    def sanitize(weights):
        sanitized_weights = {}
        for k, v in weights.items():
            if "final_logits_bias" in k:
                continue
            sanitized_weights[k] = v
        return sanitized_weights



import glob
import inspect
import json
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import mlx.core as mx
import mlx.nn as nn
import numpy as np
from mlx_lm.utils import snapshot_download

from .config import ModelConfig
from .language import LanguageModel
from .vision import VisionModel


class AyaVisionMultiModalProjector(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.downsample_factor = config.downsample_factor
        self.alignment_intermediate_size = getattr(
            config, "alignment_intermediate_size", config.text_config.hidden_size
        )
        if config.model_type == "aya_vision":
            self.layernorm = nn.LayerNorm(
                config.vision_config.hidden_size * (config.downsample_factor**2),
                eps=config.adapter_layer_norm_eps,
            )

        self.linear_1 = nn.Linear(
            config.vision_config.hidden_size * (config.downsample_factor**2),
            self.alignment_intermediate_size,
            bias=True,
        )

        self.act = nn.SiLU()  # SwiGLU uses SiLU activation

        # For SwiGLU, project down to half size since we split intermediate dim
        self.linear_2 = nn.Linear(
            self.alignment_intermediate_size // 2,
            config.text_config.hidden_size,
            bias=True,
        )

    def __call__(self, image_features):
        image_features = self.pixel_shuffle(image_features)
        if self.config.model_type == "aya_vision":
            image_features = self.layernorm(image_features)
        hidden_states = self.linear_1(image_features)

        # Split along last dimension and apply SwiGLU
        x, gate = mx.split(hidden_states, 2, axis=-1)
        hidden_states = self.act(gate) * x

        hidden_states = self.linear_2(hidden_states)
        return hidden_states

    def pixel_shuffle(self, image_features):  # B, S, D
        batch_size, seq_length, feature_dim = image_features.shape
        height = width = int(seq_length**0.5)
        image_features = image_features.reshape(
            image_features.shape[0], width, height, -1
        )
        channels = image_features.shape[-1]
        image_features = image_features.reshape(
            batch_size,
            width,
            int(height / self.downsample_factor),
            int(channels * self.downsample_factor),
        )
        image_features = image_features.transpose(0, 2, 1, 3)
        image_features = image_features.reshape(
            batch_size,
            int(height / self.downsample_factor),
            int(width / self.downsample_factor),
            -1,
        )
        image_features = image_features.transpose(0, 2, 1, 3)
        return image_features


class Model(nn.Module):
    def __init__(self, config: ModelConfig):
        super().__init__()
        self.config = config
        self.vision_tower = VisionModel(config.vision_config)
        self.language_model = LanguageModel(config.text_config)
        self.multi_modal_projector = AyaVisionMultiModalProjector(config)
        self.vision_feature_layer = config.vision_feature_layer
        self.vision_feature_select_strategy = config.vision_feature_select_strategy

    def get_input_embeddings(
        self,
        input_ids: Optional[mx.array] = None,
        pixel_values: Optional[mx.array] = None,
        **kwargs,
    ):
        if pixel_values is None:
            return self.language_model.model.embed_tokens(input_ids)

        # Get the input embeddings from the language model
        inputs_embeds = self.language_model.model.embed_tokens(input_ids)

        spatial_shapes = kwargs.get("spatial_shapes", None)
        # Get the ouptut hidden states from the vision model
        *_, hidden_states = self.vision_tower(
            pixel_values.transpose(0, 2, 3, 1),
            spatial_shapes=spatial_shapes,
            output_hidden_states=True,
        )

        # Select the hidden states from the desired layer
        selected_image_feature = hidden_states[self.vision_feature_layer]

        if self.vision_feature_select_strategy == "default":
            selected_image_feature = selected_image_feature[:, 1:]
        elif self.vision_feature_select_strategy == "full":
            selected_image_feature = selected_image_feature
        else:
            raise ValueError(
                "Unexpected feature selection strategy: "
                f"{self.vision_feature_select_strategy}"
            )

        # Pass image features through the multi-modal projector
        image_features = self.multi_modal_projector(selected_image_feature)

        # Insert special image tokens in the input_ids
        final_inputs_embeds = self._merge_input_ids_with_image_features(
            image_features, inputs_embeds, input_ids
        )
        return final_inputs_embeds

    def _merge_input_ids_with_image_features(
        self, image_features, inputs_embeds, input_ids
    ):
        image_token_index = self.config.image_token_index

        # Positions of <image> tokens in input_ids, assuming batch size is 1
        image_positions = np.where(input_ids[0] == image_token_index)[0].tolist()
        num_images, _, _, vision_hidden_size = image_features.shape

        reshaped_image_hidden_states = image_features.reshape(-1, vision_hidden_size)

        # cast to the dtype of the input_embeds to support quantized models
        reshaped_image_hidden_states = reshaped_image_hidden_states.astype(
            inputs_embeds.dtype
        )
        inputs_embeds[:, image_positions, :] = reshaped_image_hidden_states
        return inputs_embeds

    @property
    def layers(self):
        return self.language_model.model.layers

    def __call__(
        self,
        input_ids: mx.array,
        pixel_values: mx.array,
        mask: mx.array,
        cache=None,
        **kwargs,
    ):

        input_embddings = self.get_input_embeddings(input_ids, pixel_values, **kwargs)
        logits = self.language_model(
            input_ids, cache=cache, inputs_embeds=input_embddings
        )
        return logits

    def sanitize(self, weights):
        def transform_key(key):
            if "model.vision_tower" in key:
                key = key.replace("model.vision_tower", "vision_tower")
            if "model.multi_modal_projector" in key:
                key = key.replace(
                    "model.multi_modal_projector", "multi_modal_projector"
                )
            if "model.language_model" in key:
                key = key.replace("model.language_model", "language_model.model")
            if "lm_head" in key and not key.startswith("language_model"):
                key = key.replace("lm_head", "language_model.lm_head")
            return key

        return {transform_key(k): v for k, v in weights.items()}



from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from ..base import BaseModelConfig


@dataclass
class AudioConfig(BaseModelConfig):
    input_feat_size: int = 80
    hidden_size: int = 1536
    conf_attention_chunk_size: int = 12
    conf_attention_context_left: int = 13
    conf_attention_context_right: int = 0
    conf_attention_invalid_logits_value: float = -1e9
    conf_attention_logit_cap: float = 50.0
    conf_num_attention_heads: int = 8
    conf_num_hidden_layers: int = 12
    conf_conv_kernel_size: int = 5
    conf_positional_bias_size: int = 256
    conf_reduction_factor: int = 4
    conf_residual_weight: float = 0.5
    sscp_conv_channel_size: tuple[int, int] = (128, 32)
    sscp_conv_group_norm_eps: float = 1e-3
    sscp_conv_kernel_size: tuple[tuple[int, int], tuple[int, int]] = ((3, 3), (3, 3))
    sscp_conv_stride_size: tuple[tuple[int, int], tuple[int, int]] = ((2, 2), (2, 2))
    vocab_size: int = 128
    sscp_conv_eps: float = 1e-3
    rms_norm_eps: float = 1e-6
    gradient_clipping: float = 10000000000.0
    vocab_offset: int = 262_144 + 128  # text vocab size + vision vocab size


@dataclass
class VisionConfig(BaseModelConfig):
    model_type: str = "gemma3n_vision"
    num_hidden_layers: int = 12
    hidden_size: int = 2048
    intermediate_size: int = 8192
    num_attention_heads: int = 16
    patch_size: int = 16
    image_size: int = 224
    num_channels: int = 3
    rms_norm_eps: float = 1e-6
    vocab_size: int = 128
    vocab_offset: int = 262_144


@dataclass
class TextConfig(BaseModelConfig):
    model_type: str
    hidden_size: int
    num_hidden_layers: int
    intermediate_size: int
    num_attention_heads: int = 2
    head_dim: int = 256
    rms_norm_eps: float = 1.0e-6
    vocab_size: int = 262400
    vocab_size_per_layer_input: int = 262144
    num_key_value_heads: int = 4
    laurel_rank: int = 64
    frac_shared_layers: float = 0.5
    altup_active_idx: int = 0
    pad_token_id: int = 0
    altup_num_inputs: int = 4
    altup_coef_clip: Optional[float] = None
    altup_correct_scale: bool = True
    hidden_size_per_layer_input: int = 1024
    rope_local_base_freq: float = 10000.0
    rope_traditional: bool = False
    rope_theta: float = 1000000.0
    query_pre_attn_scalar: float = 0.0625
    sliding_window: int = 1024
    rope_scaling: Optional[Dict[str, Union[float, List[float]]]] = None
    mm_tokens_per_image: int = 256
    sliding_window_pattern: int = 5
    activation_sparsity_pattern: Optional[List[float]] = None
    final_logit_softcapping: float = 30.0
    query_rescale_scalar: float = 1.0
    num_kv_shared_layers: int = 0
    max_position_embeddings: int = 32768
    attn_logit_softcapping: float = 0.0
    layer_types: List[str] = None


@dataclass
class ModelConfig(BaseModelConfig):
    text_config: TextConfig
    vision_config: VisionConfig
    audio_config: AudioConfig
    model_type: str
    vocab_size: int = 257152
    ignore_index: int = -100
    image_token_index: int = 262145
    audio_token_id: int = 262273
    image_token_id: int = 262145
    hidden_size: int = 2048
    pad_token_id: int = 0
    vision_soft_tokens_per_image: int = 256
    audio_soft_tokens_per_image: int = 188
    eos_token_id: Optional[List[int]] = None


@dataclass
class MultiQueryAttentionBlockConfig(BaseModelConfig):
    num_heads: int = 8
    kv_dim: int = 16
    kv_strides: int = 1
    mmqa_avg_pool_kv: bool = False
    mmqa_dropout: float = 0.0
    mmqa_dw_kernel_size: int = 3
    is_multiscale: bool = False


@dataclass
class UniversalInvertedResidualConfig(BaseModelConfig):
    start_dw_kernel_size: int = 0  # Zero size means no conv
    mid_dw_kernel_size: int = 0  # Zero size means no conv
    filters: int = 32
    strides: int = 1
    expand_ratio: float = 4.0
    is_multiscale: bool = False


@dataclass
class EdgeResidualConfig(BaseModelConfig):
    kernel_size: int = 3
    filters: int = 32
    strides: int = 1
    expand_ratio: float = 4.0
    is_multiscale: bool = False
